"""Functions to load meshes from different file formats.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import os
import logging
import numpy as np

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import ReflectionSymmetricMesh
from capytaine.meshes.geometry import xOz_Plane, yOz_Plane
from capytaine.tools.optional_imports import import_optional_dependency

LOG = logging.getLogger(__name__)

real_str = r'[+-]?(?:\d+\.\d*|\d*\.\d+)(?:[Ee][+-]?\d+)?'  # Regex for floats


def _check_file(filename, name=None):
    if not os.path.isfile(filename):
        raise IOError("file %s not found" % filename)
    return


def load_mesh(filename, file_format=None, name=None):
    """Driver function that loads every mesh file format known by meshmagick.
    Dispatch to one of the other function depending on file_format.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk
    file_format: str, optional
        format of the mesh defined in the extension_dict dictionary
    name: str, optional
        name for the created mesh object

    Returns
    -------
    Mesh or SymmetricMesh
        the loaded mesh
    """
    _check_file(filename)

    if file_format is None:
        _, file_format = os.path.splitext(filename)
        file_format = file_format.strip('.').lower()

    if file_format not in extension_dict:
        raise IOError('Extension ".%s" is not known' % file_format)

    loader = extension_dict[file_format]

    if name is None: name = filename

    return loader(filename, name)


def load_RAD(filename, name=None):
    """Loads RADIOSS mesh files. This export file format may be chosen in ICEM meshing program.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    RAD files have a 1-indexing
    """

    import re
    _check_file(filename)
    ifile = open(filename, 'r')
    data = ifile.read()
    ifile.close()

    # node_line = r'\s*\d+(?:\s*' + real_str + '){3}'
    node_line = r'\s*\d+\s*(' + real_str + r')\s*(' + real_str + r')\s*(' + real_str + ')'
    node_section = r'((?:' + node_line + ')+)'

    elem_line = r'^\s*(?:\d+\s+){6}\d+\s*[\r\n]+'
    elem_section = r'((?:' + elem_line + '){3,})'

    pattern_node_line = re.compile(node_line, re.MULTILINE)
    # pattern_node_line_group = re.compile(node_line, re.MULTILINE)
    pattern_elem_line = re.compile(elem_line, re.MULTILINE)
    pattern_node_section = re.compile(node_section, re.MULTILINE)
    pattern_elem_section = re.compile(elem_section, re.MULTILINE)

    vertices = []
    node_section = pattern_node_section.search(data).group(1)
    for node in pattern_node_line.finditer(node_section):
        vertices.append(list(map(float, list(node.groups()))))
    vertices = np.asarray(vertices, dtype=float)

    faces = []
    elem_section = pattern_elem_section.search(data).group(1)
    for elem in pattern_elem_line.findall(elem_section):
        faces.append(list(map(int, elem.strip().split()[3:])))
    faces = np.asarray(faces, dtype=int) - 1

    return Mesh(vertices, faces, name)


def load_HST(filename, name=None):
    """Loads HYDROSTAR (Bureau Veritas (c)) mesh files.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    HST files have a 1-indexing
    """

    _check_file(filename)

    with open(filename, 'r') as f:
        lines = f.readlines()

    optional_keywords = ['PROJECT', 'SYMMETRY']
    not_implemented_optional_keywords = ['USER', 'REFLENGTH', 'GRAVITY', 'RHO', 'NBBODY']

    vertices = []
    faces = []
    optional_data = {kw: None for kw in optional_keywords}
    current_context = None
    ignored_lines = []

    for i_line, line in enumerate(lines):
        line = line.lstrip()

        if line == '':
            continue

        elif line.startswith("COORDINATES"):
            current_context = 'vertices'

        elif current_context == 'vertices' and line.startswith("ENDCOORDINATES"):
            current_context = None

        elif line.startswith("PANEL"):
            panels_type = int(line[10:])
            current_context = ('panels', panels_type)

        elif (current_context == ('panels', 0) or current_context == ('panels', 1)) and line.startswith("ENDPANEL"):
            current_context = None

        elif current_context == 'vertices':  # parse vertex coordinates
            numbers = line.split()
            if len(numbers) == 4:
                i_vertex, x, y, z = numbers
                if int(i_vertex) != len(vertices) + 1:
                    raise ValueError(
                        f"HST mesh reader expected the next vertex to be indexed as {len(vertices)+1}, "
                        f"but it was actually indexed as {i_vertex} (line {i_line+1} of {filename}).")
            elif len(numbers) == 3:
                x, y, z = numbers
            vertices.append([x, y, z])

        elif current_context == ('panels', 0):  # parse face definition (no index given)
            numbers = line.split()
            if len(numbers) == 3:
                v1, v2, v3 = numbers
                v4 = v3
            elif len(numbers) == 4:
                v1, v2, v3, v4 = numbers
            faces.append([v1, v2, v3, v4])

        elif current_context == ('panels', 1):  # parse face definition
            numbers = line.split()
            if len(numbers) == 4:
                i_face, v1, v2, v3 = numbers
                v4 = v3
            elif len(numbers) == 5:
                i_face, v1, v2, v3, v4 = numbers

            if int(i_face) != len(faces) + 1:
                ii = len(faces) + 1
                raise ValueError(f"HST mesh reader expected the next face to be indexed {ii},\n"
                                 f"but it was actually indexed with {i_face} (line {i_line+1} of file {filename}).")
            faces.append([v1, v2, v3, v4])

        elif line.startswith("ENDFILE"):
            break

        else:
            for keyword in optional_data:
                if line.startswith(keyword):
                    optional_data[keyword] = line[len(keyword)+1:].lstrip(':').strip()
                    break
            else:
                ignored_lines.append((i_line+1, line))

    if len(ignored_lines) > 0:
        formatted_ignored_lines = ["{: 4} | {}".format(i, line.strip('\n')) for (i, line) in ignored_lines]
        LOG.warning(f"HST mesh reader ignored the following lines from file {filename}:\n" + "\n".join(formatted_ignored_lines))

    vertices = np.array(vertices, dtype=float)
    faces = np.array(faces, dtype=int) - 1

    if name is None: name = optional_data['PROJECT']

    if optional_data['SYMMETRY'] == '1':
        return ReflectionSymmetricMesh(Mesh(vertices, faces, f"half_of_{name}"), xOz_Plane, name)
    elif optional_data['SYMMETRY'] == '2':
        return ReflectionSymmetricMesh(ReflectionSymmetricMesh(Mesh(vertices, faces, f"quarter_of_{name}"), yOz_Plane, f"half_of_{name}"), xOz_Plane, name)
    else:
        return Mesh(vertices, faces, name)


def load_DAT(filename, name=None):
    """Not implemented.
    Intended to load .DAT files used in DIODORE (PRINCIPIA (c))
    """
    _check_file(filename)
    raise NotImplementedError


def load_INP(filename, name=None):
    """Loads DIODORE (PRINCIPIA (c)) configuration file format.

    It parses the .INP file and extracts meshes defined in subsequent .DAT files using the different information
    contained in the .INP file.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    INP/DAT files use a 1-indexing
    """
    _check_file(filename)
    import re

    with open(filename, 'r') as f:
        text = f.read()

    # Retrieving frames into a dictionary frames
    pattern_frame_str = r'^\s*\*FRAME,NAME=(.+)[\r\n]+(.*)'
    pattern_frame = re.compile(pattern_frame_str, re.MULTILINE)

    frames = {}
    for match in pattern_frame.finditer(text):
        frame_name = match.group(1).strip()
        frame_vector = re.split(r'[, ]', match.group(2).strip())
        frames[frame_name] = np.asarray(list(map(float, frame_vector)))

    # Storing the inp layout into a list of dictionary
    pattern_node_elements = re.compile(r'^\s*\*(NODE|ELEMENT),(.*)', re.MULTILINE)
    layout = []
    mesh_files = {}
    for match in pattern_node_elements.finditer(text):
        field_dict = dict()
        field_dict['type'] = match.group(1)
        if field_dict['type'] == 'NODE':
            field_dict['INCREMENT'] = 'NO'
        opts = match.group(2).split(',')
        for opt in opts:
            key, pair = opt.split('=')
            field_dict[key] = pair.strip()

        # Retrieving information on mesh files and their usage
        file = field_dict['INPUT']
        if file in mesh_files:
            mesh_files[file][field_dict['type'] + '_CALL_INP'] += 1
        else:
            mesh_files[file] = {}
            mesh_files[file]['NODE_CALL_INP'] = 0
            mesh_files[file]['ELEMENT_CALL_INP'] = 0
            mesh_files[file][field_dict['type'] + '_CALL_INP'] += 1

        layout.append(field_dict)

        # RETRIEVING DATA SECTIONS FROM MESHFILES
        # patterns for recognition of sections
    node_line = r'\s*\d+(?:\s+' + real_str + '){3}'
    node_section = r'((?:' + node_line + ')+)'
    elem_line = r'^ +\d+(?: +\d+){3,4}[\r\n]+'  # 3 -> triangle, 4 -> quadrangle
    elem_section = r'((?:' + elem_line + ')+)'
    pattern_node_line = re.compile(node_line, re.MULTILINE)
    pattern_elem_line = re.compile(elem_line, re.MULTILINE)
    pattern_node_section = re.compile(node_section, re.MULTILINE)
    pattern_elem_section = re.compile(elem_section, re.MULTILINE)

    for file in mesh_files:
        try:
            meshfile = open(os.path.join(os.path.dirname(filename), file + '.DAT'), 'r')
        except:
            raise IOError('File {0:s} not found'.format(file + '.DAT'))
        data = meshfile.read()
        meshfile.close()

        node_section = pattern_node_section.findall(data)
        if len(node_section) > 1:
            raise IOError("""Several NODE sections into a .DAT file is not supported by meshmagick
                              as it is considered as bad practice""")
        node_array = []
        idx_array = []
        for node in pattern_node_line.findall(node_section[0]):
            node = node.split()

            node[0] = int(node[0])
            idx_array.append(node[0])
            node[1:] = list(map(float, node[1:]))
            node_array.append(node[1:])

        mesh_files[file]['NODE_SECTION'] = node_array

        # Detecting renumberings to do
        real_idx = 0
        # renumberings = []
        id_new = - np.ones(max(idx_array) + 1, dtype=int)
        # FIXME: cette partie est tres buggee !!!
        for i, idx in enumerate(idx_array):
            id_new[idx] = i+1

        mesh_files[file]['ELEM_SECTIONS'] = []
        for elem_section in pattern_elem_section.findall(data):

            elem_array = []
            for elem in pattern_elem_line.findall(elem_section):
                elem = list(map(int, elem.split()))
                # for node in elem[1:]:
                elem = id_new[elem[1:]].tolist()
                if len(elem) == 3:  # Case of a triangle, we repeat the first node at the last position
                    elem.append(elem[0])

                elem_array.append(list(map(int, elem)))
            mesh_files[file]['ELEM_SECTIONS'].append(elem_array)
        mesh_files[file]['nb_elem_sections'] = len(mesh_files[file]['ELEM_SECTIONS'])

        mesh_files[file]['nb_elem_sections_used'] = 0

    nb_nodes = 0
    nb_elems = 0
    for field in layout:
        file = field['INPUT']
        if field['type'] == 'NODE':
            nodes = np.asarray(mesh_files[file]['NODE_SECTION'], dtype=float)
            # Translation of nodes according to frame option id any
            nodes += frames[field['FRAME']]  # TODO: s'assurer que frame est une options obligatoire...

            if nb_nodes == 0:
                vertices = nodes.copy()
                nb_nodes = vertices.shape[0]
                increment = False
                continue

            if field['INCREMENT'] == 'NO':
                vertices[idx, :] = nodes.copy()
                increment = False
            else:
                vertices = np.concatenate((vertices, nodes))
                nb_nodes = vertices.shape[0]
                increment = True
        else:  # this is an ELEMENT section
            elem_section = np.asarray(mesh_files[file]['ELEM_SECTIONS'][mesh_files[file]['nb_elem_sections_used']],
                                      dtype=int)

            mesh_files[file]['nb_elem_sections_used'] += 1
            if mesh_files[file]['nb_elem_sections_used'] == mesh_files[file]['nb_elem_sections']:
                mesh_files[file]['nb_elem_sections_used'] = 0

            # Updating to new id of nodes
            elems = elem_section
            if increment:
                elems += nb_nodes

            if nb_elems == 0:
                faces = elems.copy()
                nb_elems = faces.shape[0]
                continue
            else:
                faces = np.concatenate((faces, elems))
                nb_elems = faces.shape[0]

    return Mesh(vertices, faces-1, name)


def load_TEC(filename, name=None):
    """Loads TECPLOT (Tecplot (c)) mesh files.

    It relies on the tecplot file reader from the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    TEC files have a 1-indexing
    """

    import re

    _check_file(filename)

    data_pattern = re.compile(
                    r'ZONE.*\s*N\s*=\s*(\d+)\s*,\s*E=\s*(\d+)\s*,\s*F\s*=\s*FEPOINT\s*,\s*ET\s*=\s*QUADRILATERAL\s+'
                    + r'(^(?:\s*' + real_str + r'){3,})\s+'
                    + r'(^(?:\s*\d+)*)', re.MULTILINE)

    with open(filename, 'r') as f:
        data = f.read()

    nv, nf, vertices, faces = data_pattern.search(data).groups()
    nv = int(nv)
    nf = int(nf)

    vertices = np.asarray(list(map(float, vertices.split())), dtype=float).reshape((nv, -1))[:, :3]
    faces = np.asarray(list(map(int, faces.split())), dtype=int).reshape((nf, 4))-1

    return Mesh(vertices, faces, name)


def load_VTU(filename, name=None):
    """Loads VTK file format in the new XML format (vtu file extension for unstructured meshes).

    It relies on the reader from the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    VTU files have a 0-indexing
    """

    _check_file(filename)

    vtk = import_optional_dependency("vtk")

    reader = vtk.vtkXMLUnstructuredGridReader()
    reader.SetFileName(str(filename))
    reader.Update()
    vtk_mesh = reader.GetOutput()

    vertices, faces = _dump_vtk(vtk_mesh)
    return Mesh(vertices, faces, name)


def load_VTP(filename, name=None):
    """Loads VTK file format in the new XML format (vtp file extension for polydata meshes).

    It relies on the reader from the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    VTP files have a 0-indexing
    """
    _check_file(filename)

    vtk = import_optional_dependency("vtk")

    reader = vtk.vtkXMLPolyDataReader()
    reader.SetFileName(str(filename))
    reader.Update()
    vtk_mesh = reader.GetOutput()

    vertices, faces = _dump_vtk(vtk_mesh)
    return Mesh(vertices, faces, name)


def load_VTK(filename, name=None):
    """Loads VTK file format in the legacy format (vtk file extension).

    It relies on the reader from the VTK library.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    VTU files have a 0-indexing
    """
    _check_file(filename)

    vtk = import_optional_dependency("vtk")

    reader = vtk.vtkPolyDataReader()
    reader.SetFileName(str(filename))
    reader.Update()
    vtk_mesh = reader.GetOutput()

    vertices, faces = _dump_vtk(vtk_mesh)
    return Mesh(vertices, faces, name)


def _dump_vtk(vtk_mesh):
    """Internal driver function that uses the VTK library to read VTK polydata or vtk unstructured grid data structures

    Returns
    -------
    vertices: ndarray
        numpy array of the coordinates of the mesh's nodes
    faces: ndarray
        numpy array of the faces' nodes connectivities
    """

    nv = vtk_mesh.GetNumberOfPoints()
    vertices = np.zeros((nv, 3), dtype=float)
    for k in range(nv):
        vertices[k] = np.array(vtk_mesh.GetPoint(k))

    nf = vtk_mesh.GetNumberOfCells()
    faces = np.zeros((nf, 4), dtype=int)
    for k in range(nf):
        cell = vtk_mesh.GetCell(k)
        nv_facet = cell.GetNumberOfPoints()
        for l in range(nv_facet):
            faces[k][l] = cell.GetPointId(l)
        if nv_facet == 3:
            faces[k][3] = faces[k][0]

    return vertices, faces


def load_STL(filename, name=None):
    """Loads STL file format.

    It relies on the reader from the VTK library. As STL file format maintains a redundant set of vertices for each
    faces of the mesh, it returns a merged list of nodes and connectivity array by using the merge_duplicates function.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    STL files have a 0-indexing
    """
    vtk = import_optional_dependency("vtk")

    from capytaine.meshes.quality import merge_duplicate_rows

    _check_file(filename)

    reader = vtk.vtkSTLReader()
    reader.SetFileName(str(filename))
    reader.Update()

    data = reader.GetOutputDataObject(0)

    nv = data.GetNumberOfPoints()
    vertices = np.zeros((nv, 3), dtype=float)
    for k in range(nv):
        vertices[k] = np.array(data.GetPoint(k))
    nf = data.GetNumberOfCells()
    faces = np.zeros((nf, 4), dtype=int)
    for k in range(nf):
        cell = data.GetCell(k)
        if cell is not None:
            for l in range(3):
                faces[k][l] = cell.GetPointId(l)
                faces[k][3] = faces[k][0]  # always repeating the first node as stl is triangle only

    # Merging duplicates nodes
    vertices, new_id = merge_duplicate_rows(vertices)
    faces = new_id[faces]

    return Mesh(vertices, faces, name)


def load_NAT(filename, name=None):
    """This function loads natural file format for meshes.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Notes
    -----
    The file format is as follow::

        xsym    ysym
        n    m
        x1    y1    z1
        .
        .
        .
        xn    yn    zn
        i1    j1    k1    l1
        .
        .
        .
        im    jm    km    lm

    where :
    n : number of nodes
    m : number of cells
    x1 y1 z1 : cartesian coordinates of node 1
    i1 j1 k1 l1 : counterclock wise Ids of nodes for cell 1
    if cell 1 is a triangle, i1==l1

    Note
    ----
    NAT files have a 1-indexing
    """

    _check_file(filename)

    ifile = open(filename, 'r')
    ifile.readline()
    nv, nf = list(map(int, ifile.readline().split()))

    vertices = []
    for i in range(nv):
        vertices.append(list(map(float, ifile.readline().split())))
    vertices = np.array(vertices, dtype=float)

    faces = []
    for i in range(nf):
        faces.append(list(map(int, ifile.readline().split())))
    faces = np.array(faces, dtype=int)

    ifile.close()
    return Mesh(vertices, faces-1, name)


def load_GDF(filename, name=None):
    """Loads WAMIT (Wamit INC. (c)) GDF mesh files.

    As GDF file format maintains a redundant set of vertices for each faces of the mesh, it returns a merged list of
    nodes and connectivity array by using the merge_duplicates function.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh or ReflectionSymmetricMesh
        the loaded mesh

    Note
    ----
    GDF files have a 1-indexing
    """

    _check_file(filename)

    with open(str(filename)) as gdf_file:
        title = gdf_file.readline()
        ulen, grav = map(float, gdf_file.readline().split()[:2])
        isx, isy = map(int, gdf_file.readline().split()[:2])
        npan = int(gdf_file.readline().split()[0])
        faces_vertices = np.genfromtxt(gdf_file)

    vertices, indices = np.unique(faces_vertices, axis=0, return_inverse=True)
    faces = indices.reshape(-1, 4)

    if faces.shape[0] != npan:
        raise ValueError(
            f"In {filename} npan value: {npan} is not equal to face count: \
                {faces.shape[0]}."
        )

    if isx == 1 and isy == 1:
        return ReflectionSymmetricMesh(ReflectionSymmetricMesh(Mesh(vertices, faces, f"quarter_of_{name}"), yOz_Plane, f"half_of_{name}"), xOz_Plane, name)
    elif isx == 1:
        return ReflectionSymmetricMesh(Mesh(vertices, faces, f"half_of_{name}"), yOz_Plane, name)
    elif isy == 1:
        return ReflectionSymmetricMesh(Mesh(vertices, faces, f"half_of_{name}"), xOz_Plane, name)
    else:
        return Mesh(vertices, faces, name)


def load_MAR(filename, name=None):
    """Loads Nemoh (Ecole Centrale de Nantes) mesh files.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh or ReflectionSymmetry
        the loaded mesh

    Note
    ----
    MAR files have a 1-indexing
    """

    _check_file(filename)

    ifile = open(filename, 'r')

    header = ifile.readline()
    _, symmetric_mesh = header.split()

    vertices = []
    while 1:
        line = ifile.readline()
        line = line.split()
        if line[0] == '0':
            break
        vertices.append(list(map(float, line[1:])))

    vertices = np.array(vertices, dtype=float)
    faces = []
    while 1:
        line = ifile.readline()
        line = line.split()
        if line[0] == '0':
            break
        faces.append(list(map(int, line)))

    faces = np.array(faces, dtype=int)

    ifile.close()

    if int(symmetric_mesh) == 1:
        if name is None:
            half_mesh = Mesh(vertices, faces-1)
            return ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane)
        else:
            half_mesh = Mesh(vertices, faces-1, name=f"half_of_{name}")
            return ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=name)
    else:
        return Mesh(vertices, faces-1, name)


def load_MSH(filename, name=None):
    """Loads .MSH mesh files generated by GMSH by C. Geuzaine and J.F. Remacle.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    MSH files have a 1-indexing
    """

    import re

    _check_file(filename)

    with open(filename, 'r') as file:
        data = file.read()

    nb_nodes, nodes_data = re.search(r'\$Nodes\n(\d+)\n(.+)\$EndNodes', data, re.DOTALL).groups()
    nb_elts, elts_data = re.search(r'\$Elements\n(\d+)\n(.+)\$EndElements', data, re.DOTALL).groups()

    vertices = np.asarray(list(map(float, nodes_data.split())), dtype=float).reshape((-1, 4))[:, 1:]
    vertices = np.ascontiguousarray(vertices)
    faces = []

    # Triangles
    for tri_elt in re.findall(r'(^\d+\s2(?:\s\d+)+?$)', elts_data, re.MULTILINE):
        tri_elt = list(map(int, tri_elt.split()))
        triangle = tri_elt[-3:]
        triangle.append(triangle[0])
        faces.append(triangle)

    for quad_elt in re.findall(r'(^\d+\s3(?:\s\d+)+?$)', elts_data, re.MULTILINE):
        quad_elt = list(map(int, quad_elt.split()))
        quadrangle = quad_elt[-4:]
        faces.append(quadrangle)

    faces = np.asarray(faces, dtype=int) - 1

    return Mesh(vertices, faces, name)


def load_MED(filename, name=None):
    """Loads MED mesh files generated by SALOME MECA.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    MED files have a 1-indexing
    """

    try:
        import h5py
    except ImportError:
        raise ImportError('MED file format reader needs h5py module to be installed')

    _check_file(filename)

    file = h5py.File(filename)

    list_of_names = []
    file.visit(list_of_names.append)

    nb_quadrangles = nb_triangles = 0

    for item in list_of_names:
        if '/NOE/COO' in item:
            vertices = file[item][:].reshape((3, -1)).T
            nv = vertices.shape[0]
        if '/MAI/TR3/NOD' in item:
            triangles = file[item][:].reshape((3, -1)).T - 1
            nb_triangles = triangles.shape[0]
        if '/MAI/QU4/NOD' in item:
            quadrangles = file[item][:].reshape((4, -1)).T - 1
            nb_quadrangles = quadrangles.shape[0]

    file.close()

    if nb_triangles == 0:
        triangles = np.zeros((0, 4), dtype=int)
    else:
        triangles = np.column_stack((triangles, triangles[:, 0]))
    if nb_quadrangles == 0:
        quadrangles = np.zeros((0, 4), dtype=int)

    faces = np.row_stack([triangles, quadrangles])

    return Mesh(vertices, faces, name=name)


def load_WRL(filename, name=None):
    """Loads VRML 2.0 mesh files.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh
    """
    import re

    vtk = import_optional_dependency("vtk")

    _check_file(filename)

    # Checking version
    with open(filename, 'r') as f:
        line = f.readline()
        ver = re.search(r'#VRML\s+V(\d.\d)', line).group(1)
        if not ver == '2.0':
            raise NotImplementedError('VRML loader only supports VRML 2.0 format (version %s given)' % ver)

    importer = vtk.vtkVRMLImporter()
    importer.SetFileName(str(filename))
    importer.Update()

    actors = importer.GetRenderer().GetActors()
    actors.InitTraversal()
    dataset = actors.GetNextActor().GetMapper().GetInput()

    return _dump_vtk(dataset)


def load_NEM(filename, name=None):
    """Loads mesh files that are used by the ``Mesh`` tool included in Nemoh.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh
        the loaded mesh

    Note
    ----
    This format is different from that is used directly by Nemoh software. It is only dedicated to the Mesh tool.
    """

    _check_file(filename)

    ifile = open(filename, 'r')

    nv = int(ifile.readline())
    nf = int(ifile.readline())

    vertices = []
    for ivertex in range(nv):
        vertices.append(list(map(float, ifile.readline().split())))
    vertices = np.asarray(vertices, dtype=float)

    faces = []
    for iface in range(nf):
        faces.append(list(map(int, ifile.readline().split())))
    faces = np.asarray(faces, dtype=int)
    faces -= 1

    return Mesh(vertices, faces, name)


def load_PNL(filename, name=None):
    """Load mesh using HAMS file format.

    Parameters
    ----------
    filename: str
        name of the mesh file on disk

    Returns
    -------
    Mesh or ReflectionSymmetricMesh
        the loaded mesh
    """

    with open(filename, 'r') as f:
        # Skip 3 title lines
        f.readline()
        f.readline()
        f.readline()
        # Read header data
        nb_faces, nb_vertices, x_sym, y_sym = map(int, f.readline().split())
        # Skip 2 more lines
        f.readline()
        f.readline()
        vertices = np.genfromtxt((f.readline() for _ in range(nb_vertices)), usecols=(1, 2, 3))
        # Skip 3 more lines
        f.readline()
        f.readline()
        f.readline()
        faces = np.zeros((nb_faces, 4), dtype=int)
        for i in range(nb_faces):
            index, nb_corners, *data = map(int, f.readline().split())
            assert i+1 == index
            if nb_corners == 3:  # Triangle
                assert len(data) == 3
                faces[i, 0:3] = data
                faces[i, 3] = faces[i, 2]  # Convention for triangles in Capytaine: repeat last vertex
            elif int(nb_corners) == 4:  # Quadrangle
                assert len(data) == 4
                faces[i, :] = data
    faces = faces - 1  # Going from Fortran 1-based indices to Numpy 0-based indices

    if x_sym == 1 and y_sym == 0:
        half_mesh = Mesh(vertices, faces, name=(f"half_of_{name}" if name is not None else None))
        return ReflectionSymmetricMesh(half_mesh, plane=yOz_Plane, name=name)
    elif x_sym == 0 and y_sym == 1:
        half_mesh = Mesh(vertices, faces, name=(f"half_of_{name}" if name is not None else None))
        return ReflectionSymmetricMesh(half_mesh, plane=xOz_Plane, name=name)
    elif x_sym == 1 and y_sym == 1:
        quarter_mesh = Mesh(vertices, faces, name=(f"quarter_of_{name}" if name is not None else None))
        half_mesh = ReflectionSymmetricMesh(quarter_mesh, plane=xOz_Plane, name=(f"half_of_{name}" if name is not None else None))
        return ReflectionSymmetricMesh(half_mesh, plane=yOz_Plane, name=name)
    else:
        return Mesh(vertices, faces, name)


extension_dict = {  # keyword, reader
    'dat': load_MAR,
    'mar': load_MAR,
    'nemoh': load_MAR,
    'wamit': load_GDF,
    'gdf': load_GDF,
    'diodore-inp': load_INP,
    'inp': load_INP,
    'diodore-dat': load_DAT,
    'hydrostar': load_HST,
    'hst': load_HST,
    'natural': load_NAT,
    'nat': load_NAT,
    'gmsh': load_MSH,
    'msh': load_MSH,
    'rad': load_RAD,
    'radioss': load_RAD,
    'stl': load_STL,
    'vtu': load_VTU,
    'vtp': load_VTP,
    'paraview-legacy': load_VTK,
    'vtk': load_VTK,
    'tecplot': load_TEC,
    'tec': load_TEC,
    'med': load_MED,
    'salome': load_MED,
    'vrml': load_WRL,
    'wrl': load_WRL,
    'nem': load_NEM,
    'nemoh_mesh': load_NEM,
    'pnl': load_PNL,
    'hams': load_PNL,
}



#!/usr/bin/env python
# coding: utf-8
"""Tests for the mesh submodule: definition and transformation of base meshes."""

import pytest

import numpy as np
from numpy.linalg import norm

import capytaine as cpt

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.clipper import clip
from capytaine.meshes.geometry import Plane, xOz_Plane
from capytaine.bodies.predefined import HorizontalCylinder, Sphere, Rectangle
from capytaine.meshes.predefined import mesh_rectangle

# Some meshes that will be used in the following tests.
test_mesh = Mesh(vertices=np.random.rand(4, 3), faces=[range(4)], name="test_mesh")
cylinder = HorizontalCylinder().mesh.merged()
sphere = Sphere(radius=1).mesh.merged()


def test_mesh_initialization():
    """Test how the code checks the validity of the parameters."""
    with pytest.raises(AssertionError):
        Mesh(vertices=np.random.rand(6, 3), faces=[(0, 1, 2), (3, 4, 5)])

    with pytest.raises(AssertionError):
        Mesh(vertices=np.random.rand(4, 3), faces=[(0, 1, 2, -1)])

    with pytest.raises(AssertionError):
        Mesh(vertices=np.random.rand(3, 3), faces=[(0, 1, 2, 3)])


def test_vertices_and_faces_get_and_set():
    """Test get and set functions for vertices and faces."""
    # Get faces
    faces = cylinder.faces
    vertices = cylinder.vertices

    new_cylinder = Mesh(name="new_cylinder")

    # Set faces
    with pytest.raises(AssertionError):
        new_cylinder.faces = faces
        # You should set the vertices first.
    new_cylinder.vertices = vertices
    new_cylinder.faces = faces

    assert new_cylinder == cylinder


def test_mesh_naming():
    """Test how the mesh handle names and string representation."""
    # Test automatic naming
    dummy_mesh = Mesh()  # Automatically named something like mesh_1
    other_dummy_mesh = Mesh()  # Automatically named something like mesh_2
    assert dummy_mesh.name[:5] == "mesh_"
    assert other_dummy_mesh.name[:5] == "mesh_"
    assert int(dummy_mesh.name[5:]) + 1 == int(other_dummy_mesh.name[5:])


def test_as_set_of_faces():
    """Test the representation of the mesh as a set of faces.
    Also allows to define the equality of two meshes."""
    faces = cylinder.as_set_of_faces()
    assert isinstance(faces, frozenset)
    assert len(faces) == cylinder.nb_faces
    assert all(len(face) == 4 or len(face) == 3 for face in faces)  # Each face is represented by 3 or 4 points
    assert all(len(vertex) == 3 for face in faces for vertex in face)  # Each point is represented by 3 coordinates.

    assert cylinder == cylinder  # The equality is defined as the equality of the set of faces.
    assert Mesh.from_set_of_faces(faces) == cylinder  # The mesh can be reconstructed from a set of faces.


def mesh_from_set_of_faces():
    A, B, C, D = (0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0)

    # Two triangular faces
    faces = {frozenset({A, B, C}), frozenset({B, C, D})}
    mesh = Mesh.from_set_of_faces(faces)
    assert mesh.nb_vertices == 4
    assert mesh.nb_faces == 2
    assert mesh.is_triangle(0) and mesh.is_triangle(1)


def test_copy():
    """Test the copy and renaming of meshes."""
    assert test_mesh.copy() is not test_mesh
    assert test_mesh.copy() == test_mesh
    assert test_mesh.copy(name="copy_of_the_test_mesh").name == "copy_of_the_test_mesh"
    assert test_mesh.merged() is test_mesh


def test_shape_helper_functions():
    assert np.allclose(sphere.center_of_mass_of_nodes, (0.0, 0.0, 0.0))
    assert np.isclose(sphere.diameter_of_nodes, 2.0)
    assert cylinder.squared_axis_aligned_bbox == (-5.0, 5.0, -5.0, 5.0, -5.0, 5.0)


def test_translate():
    assert cylinder.translated_x(0) == cylinder
    assert cylinder.translated_y(0) == cylinder
    assert cylinder.translated_z(0) == cylinder
    assert cylinder.translated([0, 0, 0]) == cylinder

    translated_sphere = sphere.translated([1, 2, 3])
    assert np.allclose(translated_sphere.center_of_mass_of_nodes, (1.0, 2.0, 3.0))
    assert np.isclose(translated_sphere.diameter_of_nodes, sphere.diameter_of_nodes)


def test_rotate():
    assert cylinder.rotated_x(0.0) == cylinder
    assert cylinder.rotated_y(0.0) == cylinder
    assert cylinder.rotated_z(0.0) == cylinder

    new_cylinder = cylinder.rotated_z(np.pi/2)
    assert np.allclose(new_cylinder.center_of_mass_of_nodes, cylinder.center_of_mass_of_nodes)
    assert new_cylinder.squared_axis_aligned_bbox == (-5.0, 5.0, -5.0, 5.0, -5.0, 5.0)


def test_mirror():
    new_cylinder = cylinder.mirrored(Plane())
    cylinder.mirror(Plane())
    assert new_cylinder == cylinder


def test_symmetrized():
    from capytaine.meshes.symmetric import ReflectionSymmetricMesh
    sym = cylinder.merged().symmetrized(xOz_Plane)
    assert isinstance(sym, ReflectionSymmetricMesh)


def test_mesh_quality():
    # TODO: Could be improved...
    cylinder.triangulate_quadrangles()
    cylinder.merge_duplicates(atol=1e-5)
    cylinder.heal_mesh()

def test_heal_mesh_removes_degenerate_panels():
    vertices = np.array([(0.0, 0.0, 0.0), (0.0, 1.0, 0.0), (1.0, 1.0, 0.0), (1.0, 0.0, 0.0)])
    faces = np.array([[0, 1, 2, 3], [1, 2, 2, 1]])
    mesh = cpt.Mesh(vertices, faces)
    assert mesh.nb_faces == 2
    mesh.heal_mesh()
    assert mesh.nb_faces == 1

def test_clipper():
    """Test clipping of mesh."""
    mesh = Sphere(radius=5.0, ntheta=10).mesh.merged()
    aabb = mesh.axis_aligned_bbox

    mesh.keep_immersed_part(free_surface=0.0, water_depth=np.inf)
    assert np.allclose(mesh.axis_aligned_bbox, aabb[:5] + (0,))  # the last item of the tuple has changed

    mesh.keep_immersed_part(free_surface=0.0, water_depth=1.0)
    assert np.allclose(mesh.axis_aligned_bbox, aabb[:4] + (-1, 0,))  # the last item of the tuple has changed

    # With CollectionOfMeshes (AxialSymmetry)
    mesh = Sphere(radius=5.0, ntheta=10).mesh
    aabb = mesh.merged().axis_aligned_bbox

    mesh.keep_immersed_part(free_surface=0.0, water_depth=np.inf)
    assert np.allclose(mesh.merged().axis_aligned_bbox, aabb[:5] + (0,))  # the last item of the tuple has changed

    mesh.keep_immersed_part(free_surface=0.0, water_depth=1.0)
    assert np.allclose(mesh.merged().axis_aligned_bbox, aabb[:4] + (-1, 0,))  # the last item of the tuple has changed

    # Check boundaries after clipping
    mesh = mesh_rectangle(size=(5,5), normal=(1,0,0))
    assert max([i[2] for i in mesh.immersed_part(free_surface=-1).vertices])<=-1
    assert max([i[2] for i in mesh.immersed_part(free_surface= 1).vertices])<= 1
    assert min([i[2] for i in mesh.immersed_part(free_surface=100, sea_bottom=-1).vertices])>=-1
    assert min([i[2] for i in mesh.immersed_part(free_surface=100, sea_bottom= 1).vertices])>= 1

    mesh = mesh_rectangle(size=(4,4), resolution=(1,1), normal=(1,0,0))
    tmp = list(mesh.clip(Plane(normal=(0,0.1,1),point=(0,0,-1)),inplace=False).vertices)
    tmp.sort(key=lambda x: x[2])
    tmp.sort(key=lambda x: x[1])
    assert np.allclose([i[2] for i in tmp], [-2, -0.8, -2, -1.2])


@pytest.mark.parametrize("size", [5, 6])
def test_clipper_indices(size):
    """Test clipped_mesh_faces_ids."""
    mesh = Rectangle(size=(size, size), resolution=(size, size), center=(0, 0, 0)).mesh.merged()
    clipped_mesh = clip(mesh, plane=Plane(point=(0, 0, 0), normal=(0, 0, 1)))
    faces_ids = clipped_mesh._clipping_data['faces_ids']

    assert clipped_mesh.nb_faces == len(faces_ids)
    assert all(norm(clipped_mesh.faces_centers[i] - mesh.faces_centers[face_id]) < 0.3
               for i, face_id in enumerate(faces_ids))


def test_clipper_corner_cases():
    mesh = sphere.translated_z(10.0)

    plane = Plane(point=(0, 0, 0), normal=(0, 0, 1))
    clipped_mesh = mesh.clip(plane, inplace=False)
    assert clipped_mesh == Mesh(None, None)  # Empty mesh

    plane = Plane(point=(0, 0, 0), normal=(0, 0, -1))
    clipped_mesh = mesh.clip(plane, inplace=False)
    assert clipped_mesh == mesh  # Unchanged mesh

    # Two distinct bodies
    two_spheres = Mesh.join_meshes(sphere.translated_z(10.0), sphere.translated_z(-10.0))
    plane = Plane(point=(0, 0, 0), normal=(0, 0, -1))
    one_sphere_remaining = two_spheres.clip(plane, inplace=False)
    assert one_sphere_remaining == sphere.translated_z(10.0)


def test_clipper_tolerance():
    mesh = cpt.mesh_vertical_cylinder(length=10.001, center=(0, 0, -5))
    mesh = mesh.immersed_part()
    np.testing.assert_allclose(mesh.vertices[:, 2].max(), 0.0, atol=1e-12)


def test_extract_one_face():
    i = 2
    one_face = sphere.extract_one_face(i)
    assert np.all(one_face.faces_centers[0] == sphere.faces_centers[i])



import numpy as np
import capytaine as cpt

# Solve the potential flow problems
radius = 1.0
mesh = cpt.mesh_horizontal_cylinder(radius=radius, length=1.0, center=(0, 0, 0), resolution=(5, 20, 5)).immersed_part()
body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
solver = cpt.BEMSolver()
results = solver.solve_all([cpt.RadiationProblem(body=body, omega=1.0, radiating_dof=dof) for dof in body.dofs])

# Define a set of points in the fluid on which the fluid velocities will be computed
Y, Z = np.meshgrid(np.linspace(-3.0, 3.0, 40), np.linspace(-3.0, -0.05, 20))
points = np.vstack([np.zeros(Y.size), Y.ravel(), Z.ravel()]).T
# Filter out points that are inside the floating body
points = np.array([[x, y, z] for (x, y, z) in points if y**2 + z**2 > radius**2])

# Compute the velocity field for each of the result object, that is each of the radiating dof
velocities = {}
for res in results:
    velocities[res.radiating_dof] = solver.compute_velocity( points,res)

import matplotlib.pyplot as plt
fig, axs = plt.subplots(1, 2, figsize=(12, 4))
for ax, dof in zip(axs.ravel(), ["Sway", "Heave"]):
    # Display the arrows of the velocity
    y_velocities = -velocities[dof][:, 1].real
    z_velocities = -velocities[dof][:, 2].real
    ax.quiver(points[:, 1], points[:, 2], y_velocities, z_velocities, angles="xy")
    ax.axis('equal')

    # Display the boundary of the floating body
    theta = np.linspace(-np.pi, 0.0, 100)
    ax.plot(np.cos(theta), np.sin(theta), color="red")
    ax.set_title(dof)
    ax.set_xlabel("y")
    ax.set_ylabel("z")

plt.show()



"""This module implements several classes describing matrices defined by blocks.
These matrices can be nested to recursively define Hierarchical matrices.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

from capytaine.matrices.block import BlockMatrix
from capytaine.matrices.block_toeplitz import (
    BlockToeplitzMatrix, BlockSymmetricToeplitzMatrix,
    BlockCirculantMatrix, EvenBlockSymmetricCirculantMatrix, OddBlockSymmetricCirculantMatrix,
)
from capytaine.matrices.builders import (
    cut_matrix, random_block_matrix,
    full_like, zeros_like, ones_like, identity_like,
)
from capytaine.matrices.low_rank import LowRankMatrix



import numpy as np
import pytest
import capytaine as cpt

import logging
logging.basicConfig(level=logging.DEBUG)


# DISK

def test_mesh_disk():
    from capytaine.meshes.predefined.cylinders import mesh_disk
    d = mesh_disk(resolution=(3, 6), name="foo")
    assert isinstance(d, cpt.Mesh)
    assert d.nb_faces == 18
    assert d.name == "foo"

def test_mesh_disk_position():
    from capytaine.meshes.predefined.cylinders import mesh_disk
    d = mesh_disk(center=(1.0, 0.5, -0.5), normal=(0, 1, 0))
    assert np.allclose(d.vertices[:, 1], 0.5)

def test_mesh_disk_reflection_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_disk
    d = mesh_disk(resolution=(3, 6), reflection_symmetry=True, normal=(1, 0, 0))
    assert isinstance(d, cpt.ReflectionSymmetricMesh)
    assert np.all(d.half.vertices[:, 1] >= 0.0)

def test_mesh_disk_axial_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_disk
    d = mesh_disk(resolution=(3, 6), normal=(0, 1, 0), axial_symmetry=True)
    assert isinstance(d, cpt.AxialSymmetricMesh)
    assert np.allclose(d.axis.vector, (0, 1, 0))

def test_mesh_disk_both_symmetries():
    from capytaine.meshes.predefined.cylinders import mesh_disk
    with pytest.raises(NotImplementedError):
        mesh_disk(axial_symmetry=True, reflection_symmetry=True)

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_disk_max_size(max_rad):
    from capytaine.meshes.predefined.cylinders import mesh_disk
    d = mesh_disk(radius=10.0, faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad

# VERTICAL CYLINDER

def test_mesh_vertical_cylinder():
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    v = mesh_vertical_cylinder(resolution=(2, 8, 10), name="foo")
    assert isinstance(v, cpt.Mesh)
    assert v.nb_faces == (10+2*2)*8
    assert v.name == "foo"

def test_mesh_vertical_cylinder_position():
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    v = mesh_vertical_cylinder(center=(-10.0, 5.0, 1.0), length=6.0, radius=4.0, resolution=(0, 100, 5))
    assert np.isclose(np.min(v.vertices[:, 0]), -14.0)
    assert np.isclose(np.min(v.vertices[:, 1]), 1.0)
    assert np.isclose(np.min(v.vertices[:, 2]), -2.0)

def test_mesh_vertical_cylinder_reflection_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    v = mesh_vertical_cylinder(reflection_symmetry=True, center=(5.0, 5.0, 6.0))
    assert isinstance(v, cpt.ReflectionSymmetricMesh)
    assert v.merged().nb_faces == mesh_vertical_cylinder(reflection_symmetry=False).nb_faces
    assert np.all(v.vertices[:, 2] > 0.0)

def test_mesh_vertical_cylinder_axial_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    v = mesh_vertical_cylinder(axial_symmetry=True)
    assert isinstance(v, cpt.AxialSymmetricMesh)

def test_mesh_vertical_cylinder_both_symmetries():
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    with pytest.raises(NotImplementedError):
        mesh_vertical_cylinder(axial_symmetry=True, reflection_symmetry=True)

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_vertical_max_size(max_rad):
    from capytaine.meshes.predefined.cylinders import mesh_vertical_cylinder
    d = mesh_vertical_cylinder(radius=10.0, faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad

# HORIZONTAL CYLINDER

def test_mesh_horizontal_cylinder():
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    h = mesh_horizontal_cylinder(resolution=(2, 8, 10), name="foo")
    assert isinstance(h, cpt.Mesh)
    assert h.nb_faces == (10+2*2)*8
    assert h.name == "foo"

def test_mesh_horizontal_cylinder_position():
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    h = mesh_horizontal_cylinder(length=10.0, radius=1.0, resolution=(0, 100, 3), center=(10.0, 1.0, 0.0))
    assert np.isclose(np.min(h.vertices[:, 0]), 5.0)
    assert np.isclose(np.min(h.vertices[:, 1]), 0.0)

def test_mesh_horizontal_reflection_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    h = mesh_horizontal_cylinder(reflection_symmetry=True, center=(5.0, 5.0, 5.0))
    assert isinstance(h, cpt.ReflectionSymmetricMesh)
    assert h.merged().nb_faces == mesh_horizontal_cylinder(reflection_symmetry=False).nb_faces
    assert np.all(h.vertices[:, 2] > 0.0)

def test_mesh_horizontal_translation_symmetry():
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    length = 10.0
    h = mesh_horizontal_cylinder(length=length, resolution=(0, 8, 10), translation_symmetry=True)
    assert isinstance(h, cpt.TranslationalSymmetricMesh)
    assert np.isclose(np.min(h.vertices[:, 0]), -length/2)
    assert np.isclose(np.max(h.vertices[:, 0]), length/2)

    h = mesh_horizontal_cylinder(resolution=(2, 8, 10), translation_symmetry=True)
    print(h.tree_view())
    assert isinstance(h, cpt.CollectionOfMeshes)
    assert isinstance(h[0], cpt.TranslationalSymmetricMesh)

def test_mesh_horizontal_both_symmetries():
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    h = mesh_horizontal_cylinder(resolution=(0, 8, 10), reflection_symmetry=True, translation_symmetry=True)
    assert isinstance(h, cpt.ReflectionSymmetricMesh)
    assert isinstance(h[0], cpt.TranslationalSymmetricMesh)

    h = mesh_horizontal_cylinder(resolution=(2, 8, 10), reflection_symmetry=True, translation_symmetry=True)
    assert isinstance(h, cpt.ReflectionSymmetricMesh)
    assert isinstance(h[0], cpt.CollectionOfMeshes)
    assert isinstance(h[0][0], cpt.TranslationalSymmetricMesh)
    assert isinstance(h[0][1], cpt.Mesh)

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_horizontal_max_size(max_rad):
    from capytaine.meshes.predefined.cylinders import mesh_horizontal_cylinder
    d = mesh_horizontal_cylinder(radius=10.0, faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad

# SPHERE

def test_mesh_sphere():
    from capytaine.meshes.predefined.spheres import mesh_sphere
    d = mesh_sphere(resolution=(3, 6), name="foo")
    assert isinstance(d, cpt.Mesh)
    assert d.nb_faces == 18
    assert d.name == "foo"

def test_mesh_sphere_position():
    from capytaine.meshes.predefined.spheres import mesh_sphere
    d = mesh_sphere(radius=0.1, center=(1.0, 0.5, -0.5))
    assert np.all(d.vertices[:, 0] <= 1.1)
    assert np.all(0.9 <= d.vertices[:, 0])
    assert np.all(d.vertices[:, 2] <= -0.4)
    assert np.all(-0.6 <= d.vertices[:, 2])

def test_mesh_sphere_axial_symmetry():
    from capytaine.meshes.predefined.spheres import mesh_sphere
    d = mesh_sphere(resolution=(3, 6), axial_symmetry=True)
    assert isinstance(d, cpt.AxialSymmetricMesh)
    assert np.allclose(d.axis.vector, (0, 0, 1))

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_sphere_max_size(max_rad):
    from capytaine.meshes.predefined.spheres import mesh_sphere
    d = mesh_sphere(radius=10.0, faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad
    d = mesh_sphere(
            radius=10.0, resolution=(10, 10),
            faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad

# RECTANGLES

def test_mesh_rectangle():
    from capytaine.meshes.predefined.rectangles import mesh_rectangle
    d = mesh_rectangle(resolution=(3, 6), name="foo")
    assert isinstance(d, cpt.Mesh)
    assert d.nb_faces == 18
    assert d.name == "foo"

def test_mesh_rectangle_position():
    from capytaine.meshes.predefined.rectangles import mesh_rectangle
    d = mesh_rectangle(center=(1.0, 0.5, -0.5), normal=(0, 1, 0))
    assert np.allclose(d.vertices[:, 1], 0.5)

def test_mesh_rectangle_reflection_symmetry():
    from capytaine.meshes.predefined.rectangles import mesh_rectangle
    d = mesh_rectangle(resolution=(4, 6), reflection_symmetry=True, normal=(1, 0, 0))
    assert isinstance(d, cpt.ReflectionSymmetricMesh)

def test_mesh_rectangle_both_symmetries():
    from capytaine.meshes.predefined.rectangles import mesh_rectangle
    with pytest.raises(NotImplementedError):
        mesh_rectangle(translation_symmetry=True, reflection_symmetry=True)

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_rectangle_max_size(max_rad):
    from capytaine.meshes.predefined.rectangles import mesh_rectangle
    d = mesh_rectangle(size=(10.0, 10.0), faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad

# PARALLELEPIPED

def test_mesh_parallelepiped():
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    p = mesh_parallelepiped(resolution=(3, 6, 3), name="foo")
    assert isinstance(p, cpt.Mesh)
    assert p.nb_faces == 90
    assert p.name == "foo"

def test_mesh_parallelepiped_sides():
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    p = mesh_parallelepiped(size=(5.0, 5.0, 5.0), missing_sides=["top", "bottom"])
    assert p.nb_faces == 64
    assert not np.any((abs(p.faces_centers[:, 0]) < 2.0) & (abs(p.faces_centers[:, 1]) < 2.0))
    # That is: no face with center (x, y, z) such that -2 < x < 2 and -2 < y < 2

def test_mesh_parallelepiped_position():
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    p = mesh_parallelepiped(center=(2.0, -2.0, 1.0))
    assert np.all(p.faces_centers[:, 0] <= 3.0)
    assert np.all(1.0 <= p.faces_centers[:, 0])

def test_mesh_parallelepiped_reflection_symmetry():
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    p = mesh_parallelepiped(resolution=(4, 4, 4), reflection_symmetry=True)
    assert isinstance(p, cpt.ReflectionSymmetricMesh)
    assert isinstance(p.half, cpt.ReflectionSymmetricMesh)
    assert p.nb_faces == 6*16

def test_mesh_parallelepiped_translation_symmetry():
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    p = mesh_parallelepiped(resolution=(4, 4, 4), translation_symmetry=True)
    assert isinstance(p, cpt.CollectionOfMeshes)
    assert isinstance(p[0], cpt.TranslationalSymmetricMesh)
    assert p.nb_faces == 6*16

@pytest.mark.parametrize("max_rad", [0.5, 1.0, 5.0])
def test_mesh_parallelepiped_max_size(max_rad):
    from capytaine.meshes.predefined.rectangles import mesh_parallelepiped
    d = mesh_parallelepiped(size=(10.0, 10.0, 10.0), faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad
    d = mesh_parallelepiped(size=(1.0, 10.0, 1.0), faces_max_radius=max_rad)
    assert d.faces_radiuses.max() <= max_rad



#!/usr/bin/env python

import numpy as np
import capytaine as cpt

# Initialize floating body
sphere = cpt.FloatingBody(
        mesh=cpt.mesh_sphere(
            radius=1.0,          # Dimension
            center=(0, 0, -2),   # Position
            resolution=(20, 20), # Fineness of the mesh
            ))

# DEFINE THE DOFS

# Manually defined heave,
# that is a vertical unit vector for all faces.
sphere.dofs["Heave"] = np.array(
    [(0, 0, 1) for face in sphere.mesh.faces]
)

# Bulging of the sphere,
# that is a deformation vector normal to the body at all faces.
# We can simply point to the normal vectors of the mesh for that.
sphere.dofs["Bulge"] = sphere.mesh.faces_normals

# Shearing of the sphere in the x direction.
# The deformation vector on a face is computed from the position of the face.
sphere.dofs["x-shear"] = np.array(
    [(np.cos(np.pi*z/2), 0, 0) for x, y, z in sphere.mesh.faces_centers]
)

# SOLVE DIFFRACTION PROBLEMS
solver = cpt.BEMSolver()

# Solve the problem for β=0 (incoming wave in the x direction).
problem_1 = cpt.DiffractionProblem(body=sphere, wave_direction=0, omega=1.0)
result_1 = solver.solve(problem_1)

# Solve the problem for β=π/2 (incoming wave in the y direction).
problem_2 = cpt.DiffractionProblem(body=sphere, wave_direction=np.pi/2, omega=1.0)
result_2 = solver.solve(problem_2)

# Print the generalized diffraction forces
# for the three dofs we defined
# for both values of the wave_direction β.
for result in [result_1, result_2]:
    print(f"Angle: {result.wave_direction:.2f}")
    for dof in sphere.dofs:
        force = result.forces[dof]
        print(f"{dof}: {np.abs(force):.2f}·exp({np.angle(force):.2f}i) N")
    print()



import logging

import numpy as np

from capytaine.tools.optional_imports import silently_import_optional_dependency

LOG = logging.getLogger(__name__)

# The builtin methods are stored as a list of 2D-points in [-1, 1]² and a list
# of corresponding weights. The 2D points will be remapped to the actual shape
# of the faces. They are only defined for quadrilaterals. They also work for
# triangles (although they might be subobtimal).

builtin_methods = {
        "First order": (np.array([(0.0, 0.0)]), np.array([1.0])),
        "Gauss-Legendre 2": (
            np.array([(+1/np.sqrt(3), +1/np.sqrt(3)),
             (+1/np.sqrt(3), -1/np.sqrt(3)),
             (-1/np.sqrt(3), +1/np.sqrt(3)),
             (-1/np.sqrt(3), -1/np.sqrt(3))]),
            np.array([1/4, 1/4, 1/4, 1/4])
            )
        }


def compute_quadrature_on_faces(faces, method):
    """
    Compute the quadrature points and weight for numerical integration over the faces.

    Parameters
    ----------
    faces: array of shape (nb_faces, 4, 3)
        The 3D-coordinates of each of the 4 corners of each face.
    method: string or quadpy object
        The method used to compute the quadrature scheme

    Returns
    -------
    points: array of shape (nb_faces, nb_quad_points, 3)
        The 3D-coordinates of each of the quadrature points (their number depends on the method) of each face.
    weights: array of shape (nb_faces, nb_quad_points)
        Weights associated to each of the quadrature points.
    """

    if method in builtin_methods:
        LOG.debug("Quadrature method found in builtin methods.")
        local_points, local_weights = builtin_methods[method]

    elif ((quadpy := silently_import_optional_dependency("quadpy")) is not None
                and isinstance(method, quadpy.c2._helpers.C2Scheme)):
        LOG.debug("Quadrature method is a Quadpy scheme: %s", method.name)
        local_points = method.points.T
        local_weights = method.weights

    else:
        raise ValueError(f"Unrecognized quadrature scheme: {method}.\n"
                         f"Consider using one of the following: {set(builtin_methods.keys())}")

    nb_faces = faces.shape[0]
    nb_quad_points = len(local_weights)
    points = np.empty((nb_faces, nb_quad_points, 3))
    weights = np.empty((nb_faces, nb_quad_points))

    for i_face in range(nb_faces):
        for k_quad in range(nb_quad_points):
            xk, yk = local_points[k_quad, :]
            points[i_face, k_quad, :] = (
                      (1+xk)*(1+yk) * faces[i_face, 0, :]
                    + (1+xk)*(1-yk) * faces[i_face, 1, :]
                    + (1-xk)*(1-yk) * faces[i_face, 2, :]
                    + (1-xk)*(1+yk) * faces[i_face, 3, :]
                    )/4
            dxidx = ((1+yk)*faces[i_face, 0, :] + (1-yk)*faces[i_face, 1, :]
                     - (1-yk)*faces[i_face, 2, :] - (1+yk)*faces[i_face, 3, :])/4
            dxidy = ((1+xk)*faces[i_face, 0, :] - (1+xk)*faces[i_face, 1, :]
                     - (1-xk)*faces[i_face, 2, :] + (1-xk)*faces[i_face, 3, :])/4
            detJ = np.linalg.norm(np.cross(dxidx, dxidy))
            weights[i_face, k_quad] = local_weights[k_quad] * 4 * detJ

    return points, weights



#!/usr/bin/env python
# coding: utf-8
"""Test related to the definition and use of geometric objects (planes, axes, ...)."""

import numpy as np

from capytaine.meshes.geometry import (
    e_x, e_y, e_z,
    Axis, Ox_axis, Oy_axis, Oz_axis,
    Plane, xOz_Plane, yOz_Plane, xOy_Plane,
    orthogonal_vectors, parallel_vectors,
    parallel_vectors_with_same_direction
)


def test_helper_functions():
    assert orthogonal_vectors(e_x, e_y)
    assert not orthogonal_vectors(e_x, 2*e_x)
    assert parallel_vectors(e_x, -e_x)
    assert not parallel_vectors(e_x, e_y)
    assert parallel_vectors_with_same_direction(e_y, 2*e_y)
    assert not parallel_vectors_with_same_direction(e_y, -e_y)


def test_axis():
    assert np.allclose(Axis(vector=(1, 1, 1)).vector, np.sqrt(3)/3 * np.ones((3,)))

    assert Axis(vector=(1, 1, 1), point=(0, 0, 0)) == Axis(vector=(1, 1, 1), point=(2, 2, 2))
    assert Axis(vector=(0, 0, 1), point=(0, 0, 0)) == Axis(vector=(2e-16, 3e-16, 1), point=(0,  0,  1.5))
    assert Axis(vector=(1, 1, 1), point=(0, 0, 0)) != Axis(vector=(1, 1, 1), point=(2, 2, 0))

    assert (2, 0, 0) in Ox_axis
    assert (0, 1, 0) not in Ox_axis

    assert Ox_axis.is_orthogonal_to(yOz_Plane)
    assert not Ox_axis.is_orthogonal_to((1, 1, 1))

    assert Ox_axis.angle_with_respect_to(Oy_axis) == np.pi/2
    assert Oy_axis.angle_with_respect_to(Ox_axis) == np.pi/2


def test_axis_transformation():
    assert Ox_axis.translated_x(10) == Ox_axis
    assert Ox_axis.translated_y(10) == Axis(vector=(1, 0, 0), point=(0, 10, 0))

    assert Ox_axis.rotated(Ox_axis, angle=np.pi/2) == Ox_axis
    assert Ox_axis.rotated(Oy_axis, angle=-np.pi/2) == Oz_axis

    assert Ox_axis.mirrored(plane=yOz_Plane) == Ox_axis
    assert Ox_axis.mirrored(plane=xOz_Plane.translated_y(2)) == Axis(vector=(1, 0, 0), point=(0, 4, 0))

    axis1 = Axis(vector=(1, 1, 1), point=(0, 0, 0))
    axis2 = Axis(vector=(1, 2, 3), point=(0, 0, 0))
    assert axis1.rotated_around_center_to_align_vectors(axis1.point, axis1.vector, axis2.vector) == axis2
    axis1.rotate(axis2, np.pi)
    assert axis1.rotated_around_center_to_align_vectors(axis1.point, axis1.vector, axis2.vector) == axis2
    axis1.vector *= -1
    assert axis1.rotated_around_center_to_align_vectors(axis1.point, axis1.vector, axis2.vector) == axis2

    axis1 = Axis(vector=(1, 1, 1), point=(1, 2, 0))
    axis2 = Axis(vector=(2, 2, 2), point=(0, 0, 0))
    assert axis1.translated_point_to_point(axis1.point, axis2.point) == axis2


def test_rotation_non_reference_axis():
    p = np.random.rand(3)
    axis = Axis(vector=(0, 0, 1), point=p)
    rotated_point = axis.rotate_points(p, np.pi)
    assert np.allclose(rotated_point, p)


def test_invariance_of_rotation_center():
    p = np.random.rand(3)
    axis1 = Axis(vector=(0, 0, 1), point=p)
    axis2 = Axis(vector=(0, 1, 0), point=p)
    assert np.allclose(axis1.rotated(axis2, np.pi/3).point, p)


def test_rotation_around_center_to_align_vectors_commutes_with_translation():
    n = np.random.rand(3)
    n /= np.linalg.norm(n)
    axis = Axis(vector=n, point=(0, 0, 0))
    a = axis.rotated_around_center_to_align_vectors((0, 0, 0), n, (0, 1, 0)).translated((1, 1, 1))
    b = axis.translated((1, 1, 1)).rotated_around_center_to_align_vectors((1, 1, 1), n, (0, 1, 0))
    assert np.allclose(a.vector, b.vector, (0, 1, 0))
    assert np.allclose(a.point, b.point)


def test_plane():
    assert (0, 1, 1) in yOz_Plane
    assert Oy_axis in yOz_Plane

    assert np.allclose(Plane(normal=(1, 1, 0)).normal, (np.sqrt(2)/2, np.sqrt(2)/2, 0))
    assert yOz_Plane == Plane(point=(0, 1, 1), normal=(2, 0, 0))

    assert xOy_Plane.is_orthogonal_to(Oz_axis)

    points_in_xplus = np.random.rand(10, 3) + np.array([1.0, -0.5, -0.5])
    assert np.all(yOz_Plane.distance_to_point(points_in_xplus) > 0)
    assert np.all(yOz_Plane.translated_x(-5.0).distance_to_point(points_in_xplus) > 0)
    assert not np.any(yOz_Plane.translated_x(5.0).distance_to_point(points_in_xplus) > 0)

    points_in_xminus = np.random.rand(10, 3) + np.array([-2.0, -0.5, -0.5])
    assert np.all(yOz_Plane.distance_to_point(points_in_xminus) < 0)
    assert not np.any(yOz_Plane.translated_x(-5.0).distance_to_point(points_in_xminus) < 0)
    assert np.all(yOz_Plane.translated_x(5.0).distance_to_point(points_in_xminus) < 0)


def test_plane_transformations():
    # TRANSLATIONS
    translated_plane = xOz_Plane.translate(vector=(1, 0, 0), inplace=False)
    assert xOz_Plane is not translated_plane
    assert xOz_Plane == translated_plane

    assert yOz_Plane.translated_x(10).rotated_y(np.pi/8).c == 10

    translated_plane = xOz_Plane.translate(vector=(0, 1, 0), inplace=False)
    assert translated_plane.c == 1
    assert np.all(translated_plane.normal == xOz_Plane.normal)

    # ROTATIONS
    rotated_plane = xOz_Plane.rotate(Oy_axis, angle=np.pi/12, inplace=False)
    assert rotated_plane == xOz_Plane.rotated(Oy_axis, angle=np.pi/12)
    assert xOz_Plane is not rotated_plane
    assert xOz_Plane == rotated_plane

    rotated_plane = xOz_Plane.rotate(Ox_axis, angle=np.pi/2, inplace=False)
    assert rotated_plane == xOy_Plane

    # MIRRORED BY ITSELF
    plane = Plane(normal=(1, 0, 0), point=(0.3, 0.2, 0.6))
    assert plane.mirrored(plane) != plane
    assert plane.mirrored(plane) == Plane(normal=(-1, 0, 0), point=(0.3, 0.2, 0.6))

    flipped_plane = plane.rotate(Axis(point=plane.point, vector=(0, 1, 0)), np.pi)
    assert flipped_plane == plane.mirror(plane)



#!/usr/bin/env python
# coding: utf-8

import pytest

import numpy as np
import xarray as xr

import capytaine as cpt
from capytaine import BEMSolver
from capytaine.bodies import FloatingBody
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.geometry import Axis, Plane
from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import HorizontalCylinder


def test_dof():
    nodes = np.array([[0, 0, 0], [0, 0, 1], [1, 0, 1], [1, 0, 0]])
    faces = np.array([[0, 1, 2, 3]])
    body = FloatingBody(Mesh(nodes, faces), name="one_face")
    assert body.dofs == {}

    body.add_translation_dof(direction=(1.0, 0.0, 0.0), name="1")
    assert np.allclose(body.dofs["1"], np.array([1.0, 0.0, 0.0]))

    body.add_translation_dof(direction=(0.0, 1.0, 0.0), name="2")
    assert np.allclose(body.dofs["2"], np.array([0.0, 1.0, 0.0]))

    body.add_rotation_dof(Axis(vector=(0.0, 0.0, 1.0)), name="3")
    body.add_rotation_dof(Axis(point=(0.5, 0, 0), vector=(0.0, 0.0, 1.0)), name="4")


def test_dof_name_inference():
    body = HorizontalCylinder()
    body.add_translation_dof(direction=(1, 0, 0), name="Surge_1")
    for dofname in ['Surge', 'SURGE', 'surge']:
        body.add_translation_dof(name=dofname)
        assert np.allclose(body.dofs[dofname], body.dofs['Surge_1'])

    body.add_rotation_dof(name="Pitch")
    body.add_rotation_dof(name="yaw")

    body.dofs.clear()
    body.add_all_rigid_body_dofs()


def test_rigid_body_dofs():
    import capytaine as cpt
    mesh = cpt.Sphere().mesh
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(10.0, 0.0, 0.0)))
    assert "Heave" in body.dofs
    assert np.all(body.dofs["Pitch"][:, 2] > 9.0)

def test_rigid_body_dofs_no_rotation_center_but_a_center_of_mass():
    import capytaine as cpt
    mesh = cpt.Sphere().mesh
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(), center_of_mass=(-10.0, 0.0, 0.0))
    assert np.all(body.dofs["Pitch"][:, 2] < -9.0)

def test_rigid_body_dofs_both_a_rotation_center_and_a_center_of_mass():
    import capytaine as cpt
    mesh = cpt.Sphere().mesh
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(10.0, 0.0, 0.0)),
                            center_of_mass=(-10.0, 0.0, 0.0))
    assert np.all(body.dofs["Pitch"][:, 2] > 9.0)

def test_rigid_body_dofs_neither_a_rotation_center_nor_a_center_of_mass():
    import capytaine as cpt
    mesh = cpt.Sphere().mesh
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    assert np.allclose(body._infer_rotation_center(), (0.0, 0.0, 0.0))

def test_defining_rotation_center_with_ints():
    # Issue #319
    import capytaine as cpt
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, -1)))
    body.translated_y(-2.0)

def test_healing_before_initializing_dofs():
    # Issue #367: https://github.com/capytaine/capytaine/issues/367
    # Define a mesh with a normal panel and degenerate panel
    vertices = np.array([(0.0, 0.0, 0.0), (0.0, 1.0, 0.0),
                         (1.0, 1.0, 0.0), (1.0, 0.0, 0.0)])
    faces = np.array([[0, 1, 2, 3], [1, 2, 2, 1]])
    mesh = cpt.Mesh(vertices, faces)
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    assert body.dofs["Heave"].shape[0] == body.mesh.nb_faces == 1

def test_bodies():
    body = Sphere(name="sphere", axial_symmetry=False)
    repr(body)
    assert np.allclose(body.geometric_center, (0, 0, 0))
    body.add_translation_dof(name="Surge")
    body.add_translation_dof(name="Heave")

    # Extract faces
    body.extract_faces(np.where(body.mesh.faces_centers[:, 2] < 0)[0])

    # Clipping
    body.keep_immersed_part(inplace=False)

    # Mirror of the dofs
    mirrored = body.mirrored(Plane(point=(1, 0, 0), normal=(1, 0, 0)))
    assert np.allclose(mirrored.geometric_center, np.array([2, 0, 0]))
    assert np.allclose(body.dofs['Surge'], -mirrored.dofs['Surge'])

    # Rotation of the dofs
    sideways = body.rotated(Axis(point=(0, 0, 0), vector=(0, 1, 0)), np.pi/2)
    assert np.allclose(sideways.dofs['Heave'][0], np.array([1, 0, 0]))

    upside_down = body.rotated(Axis(point=(0, 0, 0), vector=(0, 1, 0)), np.pi)
    assert np.allclose(body.dofs['Heave'], -upside_down.dofs['Heave'])

    # Copy of the body
    copy_of_body = body.copy(name="copy_of_sphere")
    copy_of_body.translate_x(10.0)
    copy_of_body.add_translation_dof(name="Heave")

    # Join bodies
    both = body.join_bodies(copy_of_body)
    assert set(both.dofs) == {'sphere__Surge', 'copy_of_sphere__Surge', 'sphere__Heave', 'copy_of_sphere__Heave'}

def test_mirror_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (0, 1, 0)
    body.mirror(cpt.xOz_Plane)
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([0, -1, 0]))

def test_translate_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (0, 0, 0)
    body.translate((0, 0, 1))
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([0, 0, 1]))

def test_rotate_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (1, 0, 0)
    body.rotate_z(np.pi)
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([-1, 0, 0]), atol=1e-10)


@pytest.mark.parametrize("z_center", [0, 2, -2])
@pytest.mark.parametrize("as_collection_of_meshes", [True, False])
def test_clipping_of_dofs(z_center, as_collection_of_meshes):
    """Check that clipping a body with a dof is the same as clipping the body ant then adding the dof."""
    full_sphere = Sphere(center=(0, 0, z_center), name="sphere", axial_symmetry=as_collection_of_meshes, clip_free_surface=False)
    axis = Axis(point=(1, 0, 0), vector=(1, 0, 0))

    full_sphere.add_rotation_dof(axis, name="test_dof")
    clipped_sphere = full_sphere.keep_immersed_part(free_surface=0.0, water_depth=np.inf, inplace=False)

    other_clipped_sphere = FloatingBody(mesh=clipped_sphere.mesh, name="other_sphere")
    other_clipped_sphere.add_rotation_dof(axis, name="test_dof")

    if clipped_sphere.mesh.nb_faces > 0:
        assert np.allclose(clipped_sphere.dofs['test_dof'], other_clipped_sphere.dofs['test_dof'])
    else:
        assert len(clipped_sphere.dofs['test_dof']) == 0

def test_cropping_body_with_manual_dof():
    # https://github.com/capytaine/capytaine/issues/204
    sphere = cpt.Sphere()
    sphere.dofs["Surge"] = [(1, 0, 0) for face in sphere.mesh.faces]
    sphere.keep_immersed_part()

def test_immersed_part():
    full_sphere = cpt.Sphere(axial_symmetry=False, name="ball")
    immersed_sphere = full_sphere.immersed_part()
    assert immersed_sphere is not full_sphere
    assert immersed_sphere.mesh == full_sphere.mesh.immersed_part()
    assert immersed_sphere.mesh.axis_aligned_bbox[5] <= 0.0
    assert immersed_sphere.name == "ball"
    full_sphere.translate_x(2.0)
    new_immersed_sphere = full_sphere.immersed_part()
    assert new_immersed_sphere is not immersed_sphere
    assert not np.allclose(new_immersed_sphere.mesh.axis_aligned_bbox, immersed_sphere.mesh.axis_aligned_bbox)


def test_mincing():
    body = HorizontalCylinder(length=10, radius=0.5, translation_symmetry=False, reflection_symmetry=False)
    body = body.minced((4, 1, 1))
    assert len(body.mesh) == 2
    assert np.all(body.mesh[0].faces_centers[:, 0] < 0)
    assert isinstance(body.mesh[0][0], Mesh)
    body = body.minced((1, 2, 2))
    assert isinstance(body.mesh[0][0][0][0], Mesh)


def test_assemble_regular_array():
    body = Sphere()
    body.add_all_rigid_body_dofs()
    array = body.assemble_regular_array(distance=2.0, nb_bodies=(2, 3))
    assert array.mesh.nb_faces == 6*body.mesh.nb_faces

    # Check name and order of the dofs
    assert list(array.dofs.keys())[0:3] == ["0_0__Surge", "0_0__Sway", "0_0__Heave"]
    assert "2_1__Heave" not in array.dofs.keys()

    # Check that the dofs corresponds to the right panels
    faces_1_0 = np.where(array.dofs["1_0__Heave"] != 0.0)[0]
    fc_1_0 = array.mesh.merged().faces_centers[faces_1_0, :]
    assert np.all(1.0 <= fc_1_0[:, 0]) and np.all(fc_1_0[:, 0] <= 3.0)  #   1 < x < 3
    assert np.all(-1.0 <= fc_1_0[:, 1]) and np.all(fc_1_0[:, 1] <= 1.0) #  -1 < y < 1


r = 1.0
locations = np.array([[1,0],[-1,0],[0,1]])*r*2
n_bodies = locations.shape[0]

@pytest.fixture
def fb_array():

    sphere = Sphere(
                    radius=r,            # Dimension
                    center=(0, 0, 0),    # Position
                    nphi=4, ntheta=10,   # Fineness of the mesh
    )
    my_axis = Axis((0, 1, 0),
                   point=(0,0,0))
    sphere.add_rotation_dof(axis=my_axis)
    sphere.keep_immersed_part()

    return sphere.assemble_arbitrary_array(locations)


def test_consistent_dofs_to_faces(fb_array):
    num_active_faces = []
    for fb_dof in fb_array.dofs.items():
        num_active_faces.append(np.count_nonzero(np.count_nonzero(fb_dof[1],axis=1)))

    ma = np.array(num_active_faces)

    tot_faces = fb_array.mesh.nb_faces
    exp_active_faces_per_dof = int(tot_faces/n_bodies)

    assert np.all(ma==exp_active_faces_per_dof)

def test_solve_hydrodynamics(fb_array):
    solver = BEMSolver()
    test_matrix = xr.Dataset(coords={
          'rho': 1e3,
          'water_depth': [np.inf],
          'omega': np.pi * 2 / 1,
          'wave_direction': 0,
          'radiating_dof': list(fb_array.dofs.keys()),
          })
    data = solver.fill_dataset(test_matrix, [fb_array],
                                 mesh=True,
                                 wavelength=True,
                                 wavenumber=True)
    assert data.influenced_dof.size == n_bodies
    assert data.radiating_dof.size == n_bodies
    assert data.added_mass.notnull().all()
    assert data.radiation_damping.notnull().all()
    assert data.diffraction_force.notnull().all()
    assert data.Froude_Krylov_force.notnull().all()


def test_cluster_bodies():
    centers = np.array([[0.0, 0.0, 0.0],
                        [10.0, 0.0, 0.0],
                        [11.0, 0.0, 0.0],
                        [10.0, 1.0, 0.0]])
    meshes = [cpt.mesh_sphere(center=c, name=str(c)) for c in centers]
    bodies = [cpt.FloatingBody(mesh=m) for m in meshes]
    joined_bodies = cpt.FloatingBody.join_bodies(*bodies)
    clustered_bodies = cpt.FloatingBody.cluster_bodies(*bodies)
    assert isinstance(clustered_bodies, cpt.FloatingBody)
    assert isinstance(clustered_bodies.mesh, cpt.CollectionOfMeshes)
    assert clustered_bodies.mesh.merged() == joined_bodies.mesh.merged()
    assert meshes[0] in clustered_bodies.mesh  # The first body is at the top level independently from the other three



# Copyright (C) 2017-2022 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>


class RigidBodyDofsPlaceholder:
    """Pass an instance of this class to the FloatingBody initializer to initialize the 6 ridig body dofs."""

    def __init__(self, rotation_center=None):
        self.rotation_center = rotation_center

    def __str__(self):
        return "RigidBodyDofsPlaceholder()"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())


def rigid_body_dofs(rotation_center=None):
    return RigidBodyDofsPlaceholder(rotation_center=rotation_center)



"""The linear solvers used in Capytaine.

They are based on numpy solvers with a thin layer for the handling of Hierarchical Toeplitz matrices.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
from scipy import linalg as sl
from scipy.sparse import linalg as ssl
from itertools import accumulate, chain

from capytaine.matrices.block import BlockMatrix
from capytaine.matrices.block_toeplitz import BlockSymmetricToeplitzMatrix, BlockCirculantMatrix

LOG = logging.getLogger(__name__)


# DIRECT SOLVER

def solve_directly(A, b):
    assert isinstance(b, np.ndarray) and A.ndim == b.ndim+1 and A.shape[-2] == b.shape[-1]
    if isinstance(A, BlockCirculantMatrix):
        LOG.debug("\tSolve linear system %s", A)
        blocks_of_diagonalization = A.block_diagonalize()
        fft_of_rhs = np.fft.fft(np.reshape(b, (A.nb_blocks[0], A.block_shape[0])), axis=0)
        try:  # Try to run it as vectorized numpy arrays.
            fft_of_result = np.linalg.solve(blocks_of_diagonalization, fft_of_rhs)
        except np.linalg.LinAlgError:  # Or do the same thing with list comprehension.
            fft_of_result = np.array([solve_directly(block, vec) for block, vec in zip(blocks_of_diagonalization, fft_of_rhs)])
        result = np.fft.ifft(fft_of_result, axis=0).reshape((A.shape[1],))
        return result

    elif isinstance(A, BlockSymmetricToeplitzMatrix):
        if A.nb_blocks == (2, 2):
            LOG.debug("\tSolve linear system %s", A)
            A1, A2 = A._stored_blocks[0, :]
            b1, b2 = b[:len(b)//2], b[len(b)//2:]
            x_plus = solve_directly(A1 + A2, b1 + b2)
            x_minus = solve_directly(A1 - A2, b1 - b2)
            return np.concatenate([x_plus + x_minus, x_plus - x_minus])/2
        else:
            # Not implemented
            LOG.debug("\tSolve linear system %s", A)
            return solve_directly(A.full_matrix(), b)

    elif isinstance(A, BlockMatrix):
        LOG.debug("\tSolve linear system %s", A)
        return solve_directly(A.full_matrix(), b)

    elif isinstance(A, np.ndarray):
        LOG.debug(f"\tSolve linear system (size: {A.shape}) with numpy direct solver.")
        return np.linalg.solve(A, b)

    else:
        raise ValueError(f"Unrecognized type of matrix to solve: {A}")


# CACHED LU DECOMPOSITION
class LUSolverWithCache:
    """Solve linear system with the LU decomposition.

    The latest LU decomposition is kept in memory, if a system with the same matrix needs to be solved again, then the decomposition is reused.

    Most of the complexity of this class comes from:
    1. @lru_cache does not work because numpy arrays are not hashable. So a basic cache system has been recoded from scratch.
    2. To be the default solver for the BasicMatrixEngine, the solver needs to support matrices for problems with one or two reflection symmetries.
    Hence, a custom way to cache the LU decomposition of the matrices involved in the direct linear resolution of the symmetric problem.
    """
    def __init__(self):
        self.cached_matrix = None
        self.cached_decomp = None

    def solve(self, A, b):
        return self.solve_with_decomp(self.cached_lu_decomp(A), b)

    def lu_decomp(self, A):
        """Return the LU decomposition of A.
        If A is BlockSymmetricToeplitzMatrix, then return a list of LU decompositions for each block of the block diagonalisation of the matrix.
        """
        if isinstance(A, BlockSymmetricToeplitzMatrix) and A.nb_blocks == (2, 2):
            A1, A2 = A._stored_blocks[0, :]
            return [self.lu_decomp(A1 + A2), self.lu_decomp(A1 - A2)]
        elif isinstance(A, np.ndarray):
            return sl.lu_factor(A)
        else:
            raise NotImplementedError("Cached LU solver is only implemented for dense matrices and 2×2 BlockSymmetricToeplitzMatrix.")

    def cached_lu_decomp(self, A):
        if not(A is self.cached_matrix):
            self.cached_matrix = A
            LOG.debug(f"Computing and caching LU decomposition")
            self.cached_decomp = self.lu_decomp(A)
        else:
            LOG.debug(f"Using cached LU decomposition")
        return self.cached_decomp

    def solve_with_decomp(self, decomp, b):
        """Solve the system using the precomputed LU decomposition.
        TODO: find a better way to differentiate a LU decomposition (returned as tuple by sl.lu_factor)
        and a set of LU decomposition (stored as a list in self.lu_decomp).
        """
        if isinstance(decomp, list):  # The matrix was a BlockSymmetricToeplitzMatrix
            b1, b2 = b[:len(b)//2], b[len(b)//2:]
            x_plus = self.solve_with_decomp(decomp[0], b1 + b2)
            x_minus = self.solve_with_decomp(decomp[1], b1 - b2)
            return np.concatenate([x_plus + x_minus, x_plus - x_minus])/2
        elif isinstance(decomp, tuple):  # The matrix was a np.ndarray
            return sl.lu_solve(decomp, b)
        else:
            raise NotImplementedError("Cached LU solver is only implemented for dense matrices and 2×2 BlockSymmetricToeplitzMatrix.")


# ITERATIVE SOLVER

class Counter:
    def __init__(self):
        self.nb_iter = 0

    def __call__(self, *args, **kwargs):
        self.nb_iter += 1


def solve_gmres(A, b):
    LOG.debug(f"Solve with GMRES for {A}.")

    if LOG.isEnabledFor(logging.INFO):
        counter = Counter()
        x, info = ssl.gmres(A, b, atol=1e-6, callback=counter)
        LOG.info(f"End of GMRES after {counter.nb_iter} iterations.")

    else:
        x, info = ssl.gmres(A, b, atol=1e-6)

    if info > 0:
        raise RuntimeError(f"No convergence of the GMRES after {info} iterations.\n"
                            "This can be due to overlapping panels or irregular frequencies.\n"
                            "In the latter case, using a direct solver can help (https://github.com/mancellin/capytaine/issues/30).")

    return x

def gmres_no_fft(A, b):
    LOG.debug(f"Solve with GMRES for {A} without using FFT.")

    x, info = ssl.gmres(A.no_toeplitz() if isinstance(A, BlockMatrix) else A, b, atol=1e-6)

    if info != 0:
        LOG.warning(f"No convergence of the GMRES. Error code: {info}")

    return x


# PRECONDITIONED SOLVER

def _block_Jacobi_coarse_corr(A, b, x0, R, RA, AcLU, DLU, diag_shapes, n):
    """
    Performs a single step of the block-Jacobi method with coarse correction.
    Can be used as a preconditioner for matrix A.

    Parameters
    ----------
    A: BlockMatrix
        System matrix
    b: array
        System right hand side vector
    x0: array
        Initial guess of the solution
    R: array
        Coarse space restriction matrix
    RA: array
        Precomputed product of R and A
    AcLU: list
        LU decomposition data of the coarse system matrix Ac (output of
        scipy.linalg.lu_factor)
    DLU: list
        List of LU decomposition data for the diagonal blocks of A
    diag_shapes: list
        List of shapes of diagonal blocks
    n: integer
        Size of the coarse problem (e.g. number of bodies simulated)

    Returns
    -------
    array
        Action of a step of the method on vector x0
    """
    # x_ps = x after the pre-smoothing step (block-Jacobi)
    x_ps = np.zeros(A.shape[0], dtype=complex)

    # the diagonal blocks of A have already been put to zero in build_matrices
    # they are not needed anymore
    q = b - A@x0
    # loop over diagonal blocks
    for kk in range(n):
        local_slice = slice(sum(diag_shapes[:kk]), sum(diag_shapes[:kk+1]))
        local_rhs = q[local_slice]
        local_sol = sl.lu_solve(DLU[kk], local_rhs, check_finite=False)

        x_ps[local_slice] = local_sol

    r_c = R@b - RA@x_ps #restricted residual
    e_c = sl.lu_solve(AcLU, r_c, check_finite=False)
    # update
    return x_ps + R.T@e_c

def solve_precond_gmres(A_and_precond_data, b):
    """
    Implementation of the preconditioner presented in
    `<https://doi.org/10.1007/978-3-031-50769-4_14>`.
    """
    A, R, RA, AcLU, DLU, diag_shapes, n, PinvA = A_and_precond_data
    N = A.shape[0]

    Pinvb = _block_Jacobi_coarse_corr(A, b, np.zeros(N, dtype=complex), R, RA, AcLU, DLU, diag_shapes, n)

    LOG.debug(f"Solve with GMRES for {A}.")

    if LOG.isEnabledFor(logging.INFO):
        counter = Counter()
        x, info = ssl.gmres(PinvA, Pinvb, atol=1e-6, callback=counter)
        LOG.info(f"End of GMRES after {counter.nb_iter} iterations.")

    else:
        x, info = ssl.gmres(PinvA, Pinvb, atol=1e-6)

    if info > 0:
        raise RuntimeError(f"No convergence of the GMRES after {info} iterations.\n"
                            "This can be due to overlapping panels or irregular frequencies.\n"
                            "In the latter case, using a direct solver can help (https://github.com/mancellin/capytaine/issues/30).")

    return x



#!/usr/bin/env python
# coding: utf-8
"""Tests for the computation of the Green function using Fortran routines."""

import pytest
from pytest import approx

import numpy as np
import capytaine as cpt



# from numpy import pi
# from scipy.integrate import quad
# from scipy.misc import derivative
# from scipy.optimize import newton
# from scipy.special import exp1
# from hypothesis import given

# def E1(z):
#     return np.exp(-z)*Delhommeau_f90.initialize_green_wave.exp_e1(z)
#
# @given(floats(min_value=-1e2, max_value=-1e-2), floats(min_value=-1e2, max_value=1e2))
# def test_exponential_integral(x, y):
#     z = x + 1j*y
#
#     # Compare with Scipy implementation
#     assert np.isclose(E1(z), exp1(z), rtol=1e-3)
#
#     # Test property (A3.5) of the function according to [Del, p.367].
#     if y != 0.0:
#         assert np.isclose(E1(np.conjugate(z)), np.conjugate(E1(z)), rtol=1e-3)
#
#     # BROKEN...
#     # def derivative_of_complex_function(f, z, **kwargs):
#     #     direction = 1
#     #     return derivative(lambda eps: f(z + eps*direction), 0.0, **kwargs)
#     # if abs(x) > 1 and abs(y) > 1:
#     #     assert np.isclose(
#     #         derivative_of_complex_function(Delhommeau_f90.initialize_green_wave.gg, z),
#     #         Delhommeau_f90.initialize_green_wave.gg(z) - 1.0/z,
#     #         atol=1e-2)
#

# CHECK OF THE THEORY, NOT THE CODE ITSELF
# Not necessary to run every time...
#
# @given(r=floats(min_value=0, max_value=1e2),
#        z=floats(min_value=-16, max_value=-1),
#        k=floats(min_value=0.1, max_value=10)
#        )
# def test_zeta(r, z, k):
#     """Check expression k/π ∫ 1/ζ(θ) dθ = - 1/r """

#     def one_over_zeta(theta):
#         return np.real(1/(k*(z + 1j*r*np.cos(theta))))

#     int_one_over_zeta, *_ = quad(one_over_zeta, -pi/2, pi/2)

#     assert np.isclose(k/pi * int_one_over_zeta,
#                       -1/(np.sqrt(r**2 + z**2)),
#                       rtol=1e-4)


gfs = [
        cpt.Delhommeau(tabulation_nr=328, tabulation_nz=46, tabulation_nb_integration_points=251, tabulation_method="legacy"),
        cpt.XieDelhommeau(tabulation_nr=328, tabulation_nz=46, tabulation_nb_integration_points=251, tabulation_method="legacy"),
        ]

def test_compare_tabulations_of_Delhommeau_and_XieDelhommeau():
    assert np.allclose(gfs[0].tabulated_integrals[:, :, 0, 0],
                       gfs[1].tabulated_integrals[:, :, 0, 0])
    assert np.allclose(gfs[0].tabulated_integrals[:, :, 1, 0],
                       gfs[1].tabulated_integrals[:, :, 1, 0])
    assert np.allclose(gfs[0].tabulated_integrals[:, :, 1, 1],
                       gfs[1].tabulated_integrals[:, :, 1, 1])

    r_range = gfs[0].tabulated_r_range
    z_range = gfs[0].tabulated_z_range

    # Make 2D arrays
    r = r_range[:, None] * np.ones_like(z_range)[None, :]
    z = np.ones_like(r_range)[:, None] * z_range[None, :]
    R1 = np.hypot(r, z)

    # Compare Z1, for great values of Z, where both methods are as accurate
    Del = gfs[0].tabulated_integrals[:, :, 0, 1] - 1/R1
    Xie = gfs[1].tabulated_integrals[:, :, 0, 1]
    assert np.allclose(Del[abs(z) > 1], Xie[abs(z) > 1], atol=1e-3)


@pytest.mark.parametrize("gf", gfs)
def test_symmetry_of_the_green_function_infinite_depth(gf):
    k = 1.0
    xi = np.array([0.0, 0.0, -1.0])
    xj = np.array([1.0, 1.0, -2.0])
    g1, dg1_sym, dg1_antisym = gf.fortran_core.green_wave.wave_part_infinite_depth(xi, xj, k, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals)
    g2, dg2_sym, dg2_antisym = gf.fortran_core.green_wave.wave_part_infinite_depth(xj, xi, k, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals)
    assert g1 == approx(g2)
    assert dg1_sym == approx(dg2_sym)
    assert dg1_antisym == approx(-dg2_antisym)


@pytest.mark.parametrize("gf", gfs)
def test_symmetry_of_the_green_function_finite_depth_no_prony(gf):
    k = 1.0
    depth = 5.0
    xi = np.array([0.0, 0.0, -1.0])
    xj = np.array([1.0, 1.0, -2.0])
    g1, dg1_sym, dg1_antisym = gf.fortran_core.green_wave.wave_part_finite_depth(xi, xj, k, depth, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals, np.zeros(1), np.zeros(1), 1)
    g2, dg2_sym, dg2_antisym = gf.fortran_core.green_wave.wave_part_finite_depth(xj, xi, k, depth, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals, np.zeros(1), np.zeros(1), 1)
    assert g1 == approx(g2)
    assert dg1_sym == approx(dg2_sym)
    assert dg1_antisym == approx(-dg2_antisym)


@pytest.mark.parametrize("gf", gfs)
def test_symmetry_of_the_green_function_finite_depth(gf):
    k = 1.0
    depth = 10.0
    xi = np.array([0.0, 0.0, -1.0])
    xj = np.array([1.0, 1.0, -2.0])
    ambda, a, nexp = gf.fortran_core.old_prony_decomposition.lisc(k*depth*np.tanh(k*depth), k*depth)
    g1, dg1_sym, dg1_antisym = gf.fortran_core.green_wave.wave_part_finite_depth(xi, xj, k, depth, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals, ambda, a, 31)
    g2, dg2_sym, dg2_antisym = gf.fortran_core.green_wave.wave_part_finite_depth(xj, xi, k, depth, gf.tabulation_method_index, gf.tabulated_r_range, gf.tabulated_z_range, gf.tabulated_integrals, ambda, a, 31)
    assert g1 == approx(g2)
    assert dg1_sym == approx(dg2_sym)
    assert dg1_antisym == approx(-dg2_antisym)


def test_floating_point_precision():
    assert cpt.Delhommeau(floating_point_precision="float64").tabulated_integrals.dtype == np.float64
    assert cpt.Delhommeau(floating_point_precision="float32").tabulated_integrals.dtype == np.float32


def test_no_tabulation():
    mesh = cpt.Sphere().mesh.keep_immersed_part()
    tabed_gf = cpt.Delhommeau()
    untabed_gf = cpt.Delhommeau(tabulation_nr=0, tabulation_nz=0)
    assert np.allclose(untabed_gf.evaluate(mesh, mesh)[0], tabed_gf.evaluate(mesh, mesh)[0], atol=1e-2)



"""Computing the potential and velocity of Airy wave."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import numpy as np
from capytaine.tools.lists_of_points import _normalize_points, _normalize_free_surface_points

def airy_waves_potential(points, pb):
    """Compute the potential for Airy waves at a given point (or array of points).

    Parameters
    ----------
    points: array of shape (3) or (N x 3)
        coordinates of the points in which to evaluate the potential.
    pb: DiffractionProblem
        problem with the environmental conditions (g, rho, ...) of interest

    Returns
    -------
    array of shape (1) or (N x 1)
        The potential
    """
    points, output_shape = _normalize_points(points)

    x, y, z = points.T
    k = pb.wavenumber
    h = pb.water_depth
    beta = pb.encounter_wave_direction
    wbar = x * np.cos(beta) + y * np.sin(beta)

    if 0 <= k*h < 20:
        cih = np.cosh(k*(z+h))/np.cosh(k*h)
        # sih = np.sinh(k*(z+h))/np.cosh(k*h)
    else:
        cih = np.exp(k*z)
        # sih = np.exp(k*z)

    phi = -1j*pb.g/pb.omega * cih * np.exp(1j * k * wbar)
    return phi.reshape(output_shape)


def airy_waves_velocity(points, pb):
    """Compute the fluid velocity for Airy waves at a given point (or array of points).

    Parameters
    ----------
    points: array of shape (3) or (N x 3)
        coordinates of the points in which to evaluate the potential.
    pb: DiffractionProblem
        problem with the environmental conditions (g, rho, ...) of interest

    Returns
    -------
    array of shape (3) or (N x 3)
        the velocity vectors
    """

    points, output_shape = _normalize_points(points)

    x, y, z = points.T
    k = pb.wavenumber
    h = pb.water_depth
    beta = pb.encounter_wave_direction

    wbar = x * np.cos(beta) + y * np.sin(beta)

    if 0 <= k*h < 20:
        cih = np.cosh(k*(z+h))/np.cosh(k*h)
        sih = np.sinh(k*(z+h))/np.cosh(k*h)
    else:
        cih = np.exp(k*z)
        sih = np.exp(k*z)

    v = pb.g*k/pb.omega * \
        np.exp(1j * k * wbar) * \
        np.array([np.cos(pb.wave_direction) * cih, np.sin(pb.wave_direction) * cih, -1j * sih])

    return v.T.reshape((*output_shape, 3))


def airy_waves_pressure(points, pb):
    return 1j * pb.omega * pb.rho * airy_waves_potential(points, pb)


def froude_krylov_force(pb):
    return pb.body.integrate_pressure(airy_waves_pressure(pb.body.mesh.faces_centers, pb))


def airy_waves_free_surface_elevation(points, pb):
    """Compute the free surface elevation at points of the undisturbed Airy waves

    Parameters
    ----------
    points: array of shape (3) or (N × 3) or (2) or (N × 2)
        coordinates of the points in which to evaluate the potential.
        If only two coordinates are passed, the last one is filled with zeros.
    pb: DiffractionProblem
        problem with the environmental conditions (g, rho, ...) of interest

    Returns
    -------
    complex-valued array of shape (1,) or (N,)
        the free surface elevations
    """
    points, output_shape = _normalize_free_surface_points(points)
    return 1j * pb.omega / pb.g * airy_waves_potential(points, pb).reshape(output_shape)



from numpy import pi

import capytaine as cpt
from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
from capytaine.ui.vtk import Animation

cpt.set_logging('INFO')

bem_solver = cpt.BEMSolver()


def generate_boat():
    boat_mesh = cpt.load_mesh("boat_200.mar", file_format="mar")
    boat = cpt.FloatingBody(
            mesh=boat_mesh,
            dofs=cpt.rigid_body_dofs(rotation_center=boat_mesh.center_of_buoyancy),
            center_of_mass = boat_mesh.center_of_buoyancy,
            name="pirate ship"
            )
    boat.inertia_matrix = boat.compute_rigid_body_inertia() / 10 # Artificially lower to have a more appealing animation
    boat.hydrostatic_stiffness = boat.immersed_part().compute_hydrostatic_stiffness()
    return boat


def setup_animation(body, fs, omega, wave_amplitude, wave_direction):
    # SOLVE BEM PROBLEMS
    radiation_problems = [cpt.RadiationProblem(omega=omega, body=body.immersed_part(), radiating_dof=dof) for dof in body.dofs]
    radiation_results = bem_solver.solve_all(radiation_problems)
    diffraction_problem = cpt.DiffractionProblem(omega=omega, body=body.immersed_part(), wave_direction=wave_direction)
    diffraction_result = bem_solver.solve(diffraction_problem)

    dataset = cpt.assemble_dataset(radiation_results + [diffraction_result])
    rao = cpt.post_pro.rao(dataset, wave_direction=wave_direction)

    # COMPUTE FREE SURFACE ELEVATION
    # Compute the diffracted wave pattern
    incoming_waves_elevation = airy_waves_free_surface_elevation(fs, diffraction_result)
    diffraction_elevation = bem_solver.compute_free_surface_elevation(fs, diffraction_result)

    # Compute the wave pattern radiated by the RAO
    radiation_elevations_per_dof = {res.radiating_dof: bem_solver.compute_free_surface_elevation(fs, res) for res in radiation_results}
    radiation_elevation = sum(rao.sel(omega=omega, radiating_dof=dof).data * radiation_elevations_per_dof[dof] for dof in body.dofs)

    # SET UP ANIMATION
    # Compute the motion of each face of the mesh for the animation
    rao_faces_motion = sum(rao.sel(omega=omega, radiating_dof=dof).data * body.dofs[dof] for dof in body.dofs)

    # Set up scene
    animation = Animation(loop_duration=2*pi/omega)
    animation.add_body(body, faces_motion=wave_amplitude*rao_faces_motion)
    animation.add_free_surface(fs, wave_amplitude * (incoming_waves_elevation + diffraction_elevation + radiation_elevation))
    return animation


if __name__ == '__main__':
    body = generate_boat()
    fs = cpt.FreeSurface(x_range=(-100, 75), y_range=(-100, 75), nx=100, ny=100)

    anim = setup_animation(body, fs, omega=1.5, wave_amplitude=0.5, wave_direction=pi)
    anim.run(camera_position=(70, 70, 100), resolution=(800, 600))
    anim.save("animated_boat.ogv", camera_position=(70, 70, 100), resolution=(800, 600))



"""This module contains some helpful functions to create block matrices."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from itertools import accumulate

import numpy as np

from capytaine.matrices.block import BlockMatrix
from capytaine.matrices.low_rank import LowRankMatrix

LOG = logging.getLogger(__name__)


def cut_matrix(full_matrix, x_shapes, y_shapes, check=False):
    """Transform a numpy array into a block matrix of numpy arrays.

    Parameters
    ----------
    full_matrix: numpy array
        The matrix to split into blocks.
    x_shapes: sequence of int
        The columns at which to split the blocks.
    y_shapes: sequence of int
        The lines at which to split the blocks.
    check: bool, optional
        Check to dimensions and type of the matrix after creation (default: False).

    Return
    ------
    BlockMatrix
        The same matrix as the input one but in block form.
    """
    new_block_matrix = []
    for i, di in zip(accumulate([0] + x_shapes[:-1]), x_shapes):
        line = []
        for j, dj in zip(accumulate([0] + x_shapes[:-1]), y_shapes):
            line.append(full_matrix[i:i+di, j:j+dj])
        new_block_matrix.append(line)
    return BlockMatrix(new_block_matrix, check=check)


def random_block_matrix(x_shapes, y_shapes, rng=np.random.default_rng()):
    """A random block matrix."""
    return cut_matrix(rng.uniform(size=(sum(x_shapes), sum(y_shapes))), x_shapes, y_shapes)


def full_like(A, value, dtype=np.float64):
    """A matrix of the same kind and shape as A but filled with a single value."""
    if isinstance(A, BlockMatrix):
        new_matrix = []
        for i in range(A._stored_nb_blocks[0]):
            line = []
            for j in range(A._stored_nb_blocks[1]):
                line.append(full_like(A._stored_blocks[i, j], value, dtype=dtype))
            new_matrix.append(line)
        return A.__class__(new_matrix)
    elif isinstance(A, LowRankMatrix):
        return LowRankMatrix(np.ones((A.shape[0], 1)), np.full((1, A.shape[1]), value))
    elif isinstance(A, np.ndarray):
        return np.full_like(A, value, dtype=dtype)


def zeros_like(A, dtype=np.float64):
    """A matrix of the same kind and shape as A but filled with zeros."""
    return full_like(A, 0.0, dtype=dtype)


def ones_like(A, dtype=np.float64):
    """A matrix of the same kind and shape as A but filled with ones."""
    return full_like(A, 1.0, dtype=dtype)


def identity_like(A, dtype=np.float64):
    """A identity matrix of the same kind and shape as A."""
    if isinstance(A, BlockMatrix):
        I = []
        for i in range(A._stored_nb_blocks[0]):
            line = []
            for j in range(A._stored_nb_blocks[1]):
                if i == j:
                    line.append(identity_like(A._stored_blocks[i, j], dtype=dtype))
                else:
                    line.append(zeros_like(A._stored_blocks[i, j], dtype=dtype))
            I.append(line)
        return A.__class__(I)
    elif isinstance(A, np.ndarray):
        return np.eye(A.shape[0], A.shape[1], dtype=dtype)



"""This module implements a tools to clip meshes against a plane.
Based on meshmagick <https://github.com/LHEEA/meshmagick> by François Rongère.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin, based on the work of François Rongère
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np

from capytaine.meshes.geometry import Plane
from capytaine.meshes.meshes import Mesh

LOG = logging.getLogger(__name__)


def clip(source_mesh: Mesh, plane: Plane, vicinity_tol=1e-12, name=None):
    """Return a new mesh containing the source mesh clipped by the plane.

    Parameters
    ----------
    source_mesh : Mesh
        The mesh to be clipped.
    plane : Plane, optional
        The clipping plane.
    vicinity_tol : float, optional
        The absolute tolerance to consider en vertex is on the plane. Default is 1e-12.
    name: string, optional
        A name for the new clipped mesh.
    """
    vertices_data = _vertices_positions_wrt_plane(source_mesh, plane, vicinity_tol)

    nb_vertices_above_or_on_plane = np.count_nonzero(
        vertices_data['vertices_above_mask'] | vertices_data['vertices_on_mask']
    )
    nb_vertices_below_or_on_plane = np.count_nonzero(
        vertices_data['vertices_below_mask'] | vertices_data['vertices_on_mask']
    )

    if nb_vertices_above_or_on_plane == source_mesh.nb_vertices:
        LOG.warning(f"Clipping {source_mesh.name} by {plane}: all vertices are removed.")
        clipped_mesh = Mesh(None, None)
        clipped_mesh._clipping_data = dict(faces_ids=[])

    elif nb_vertices_below_or_on_plane == source_mesh.nb_vertices:
        LOG.info(f"Clipping {source_mesh.name} by {plane}: no action.")
        clipped_mesh = source_mesh.copy()
        clipped_mesh._clipping_data = dict(faces_ids=list(range(source_mesh.nb_faces)))

    else:
        upper_mesh, crown_mesh, lower_mesh = _partition_mesh(vertices_data, source_mesh)

        if crown_mesh.nb_faces > 0:
            clipped_crown_mesh = _clip_crown(crown_mesh, plane)
            clipped_mesh = lower_mesh + clipped_crown_mesh
            clipped_mesh._clipping_data = {
                'faces_ids': np.concatenate((lower_mesh._clipping_data['faces_ids'],
                                             clipped_crown_mesh._clipping_data['faces_ids']))
            }
        else:
            clipped_mesh = lower_mesh

    if name is None:
        clipped_mesh.name = f'{source_mesh.name}_clipped'
    clipped_mesh.remove_unused_vertices()

    return clipped_mesh


def _vertices_positions_wrt_plane(source_mesh, plane, vicinity_tol):
    """Classifies vertices with respect to the clipping plane."""
    vertices_distances = plane.distance_to_point(source_mesh.vertices)
    vertices_data = {'vertices_distances': vertices_distances,
                     'vertices_above_mask': vertices_distances > vicinity_tol,
                     'vertices_on_mask': np.abs(vertices_distances) < vicinity_tol,
                     'vertices_below_mask': vertices_distances < -vicinity_tol
                     }
    return vertices_data


def _partition_mesh(vertices_data, source_mesh):
    """Partitions the mesh in 3 with respect to the plane:
    * upper_mesh: part entirely above the clipping plane
    * crown_mesh: part intersecting the clipping plane
    * lower_mesh: part entirely under the clipping plane
    """
    nb_vertices_above_per_face = vertices_data['vertices_above_mask'][source_mesh.faces].sum(axis=1)
    nb_vertices_below_per_face = vertices_data['vertices_below_mask'][source_mesh.faces].sum(axis=1)

    # Simple criteria ensuring that _faces are totally above or below the plane (4 _vertices at the same side)
    # Works for both triangles and quadrangles
    above_faces_mask = nb_vertices_above_per_face == 4
    below_faces_mask = nb_vertices_below_per_face == 4
    crown_faces_mask = np.logical_not(np.logical_or(above_faces_mask, below_faces_mask))

    faces_ids = {
        'upper_mesh': np.where(above_faces_mask)[0],
        'crown_mesh': np.where(crown_faces_mask)[0],
        'lower_mesh': np.where(below_faces_mask)[0]
    }

    partition = []
    for name in ['upper_mesh', 'crown_mesh', 'lower_mesh']:
        new_mesh, vertices_ids = source_mesh.extract_faces(
            id_faces_to_extract=faces_ids[name], return_index=True, name=name)
        new_mesh._clipping_data = {
            'faces_ids': faces_ids[name],
            'vertices_ids': vertices_ids,
            'vertices_distances': vertices_data['vertices_distances'][vertices_ids],
            'above_vertices_mask': vertices_data['vertices_above_mask'][vertices_ids],
            'on_vertices_mask': vertices_data['vertices_on_mask'][vertices_ids],
            'below_vertices_mask': vertices_data['vertices_below_mask'][vertices_ids],
        }
        partition.append(new_mesh)

    return partition


def _clip_crown(crown_mesh, plane):
    """Performs the clipping operation of the crown_mesh and determines the obtained boundaries.
    This is the heart method of the class.
    """
    vertices = crown_mesh.vertices

    above_vertices_mask = crown_mesh._clipping_data['above_vertices_mask']
    below_vertices_mask = crown_mesh._clipping_data['below_vertices_mask']

    on_vertices_mask = crown_mesh._clipping_data['on_vertices_mask']
    # TODO: Vertices pre-projection to be done here

    vertices_distances = crown_mesh._clipping_data['vertices_distances']

    # Init
    clipped_crown_mesh_faces = list()
    clipped_crown_mesh_relative_faces_ids = list()

    direct_boundary_edges = dict()
    inv_boundary_edges = dict()
    intersections_vertices = list()

    index_new_vertices = crown_mesh.nb_vertices

    for face_id in range(crown_mesh.nb_faces):

        face = crown_mesh.get_face(face_id)

        # # Determining the type of face clipping
        v_above_face = np.where(above_vertices_mask[face])[0]
        v_on_face = np.where(on_vertices_mask[face])[0]
        v_below_face = np.where(below_vertices_mask[face])[0]

        nb_above = len(v_above_face)
        nb_on = len(v_on_face)
        nb_below = len(v_below_face)

        face_type = str(nb_above) + str(nb_on) + str(nb_below)

        if face_type == '202':
            #    0*-----*3
            #     |     |
            # ----o-----o----
            #     |     |
            #    1*-----*2
            if v_above_face[1] == v_above_face[0] + 1:
                face = np.roll(face, -v_above_face[1])
            p0, p1, p2, p3 = vertices[face]
            ileft = plane.get_edge_intersection(p0, p1)
            iright = plane.get_edge_intersection(p2, p3)
            intersections_vertices += [ileft, iright]
            boundary_edge = [index_new_vertices, index_new_vertices + 1]
            clipped_crown_mesh_faces.append([index_new_vertices, face[1], face[2], index_new_vertices + 1])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 2

        elif face_type == '301':
            #      *2
            #     / \
            #    /   \
            #   /     \
            # 3*       *1
            #   \     /
            # ---o---o---
            #     \ /
            #      *0
            face = np.roll(face, -v_below_face[0])
            p0, p1, p3 = vertices[face[[0, 1, 3]]]
            ileft = plane.get_edge_intersection(p0, p3)
            iright = plane.get_edge_intersection(p0, p1)
            intersections_vertices += [ileft, iright]
            boundary_edge = [index_new_vertices, index_new_vertices + 1]
            clipped_crown_mesh_faces.append([index_new_vertices, face[0], index_new_vertices + 1, index_new_vertices])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 2

        elif face_type == '103':
            #      *0
            #     / \
            # ---o---o---
            #   /     \
            # 1* - - - *3
            #   \     /
            #    \   /
            #     \ /
            #      *2
            face = np.roll(face, -v_above_face[0])
            p0, p1, p3 = vertices[face[[0, 1, 3]]]
            ileft = plane.get_edge_intersection(p0, p1)
            iright = plane.get_edge_intersection(p0, p3)
            intersections_vertices += [ileft, iright]
            boundary_edge = [index_new_vertices, index_new_vertices + 1]
            clipped_crown_mesh_faces.append([index_new_vertices, face[1], face[3], index_new_vertices + 1])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            clipped_crown_mesh_faces.append([face[1], face[2], face[3], face[1]])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 2

        elif face_type == '102':
            #      *O
            #     / \
            # ---o---o---
            #   /     \
            # 1*-------*2
            face = np.roll(face, -v_above_face[0])
            p0, p1, p2 = vertices[face]
            ileft = plane.get_edge_intersection(p0, p1)
            iright = plane.get_edge_intersection(p0, p2)
            intersections_vertices += [ileft, iright]
            boundary_edge = [index_new_vertices, index_new_vertices + 1]
            clipped_crown_mesh_faces.append([index_new_vertices, face[1], face[2], index_new_vertices + 1])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 2

        elif face_type == '201':  # done
            #  2*-------*1
            #    \     /
            #  ---o---o---
            #      \ /
            #       *0
            face = np.roll(face, -v_below_face[0])
            p0, p1, p2 = vertices[face]
            ileft = plane.get_edge_intersection(p0, p2)
            iright = plane.get_edge_intersection(p0, p1)
            intersections_vertices += [ileft, iright]
            boundary_edge = [index_new_vertices, index_new_vertices + 1]
            clipped_crown_mesh_faces.append([index_new_vertices, face[0], index_new_vertices + 1, index_new_vertices])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 2

        elif face_type == '211':
            #        *3                   *1
            #       / \                  / \
            #      /   *2       or     2*   \
            #    0/   /                  \   \0
            # ---*---o---              ---o---*---
            #     \ /                      \ /
            #      *1                       *3
            #

            face = np.roll(face, -v_on_face[0])
            if vertices_distances[face[1]] < 0.:
                p1, p2 = vertices[face[[1, 2]]]
                iright = plane.get_edge_intersection(p1, p2)
                intersections_vertices.append(iright)
                boundary_edge = [face[0], index_new_vertices]
                clipped_crown_mesh_faces.append([face[0], face[1], index_new_vertices, face[0]])
            else:
                p2, p3 = vertices[face[[2, 3]]]
                ileft = plane.get_edge_intersection(p2, p3)
                intersections_vertices.append(ileft)
                boundary_edge = [index_new_vertices, face[0]]
                clipped_crown_mesh_faces.append([index_new_vertices, face[3], face[0], index_new_vertices])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 1

        elif face_type == '112':
            #       *3                     *1
            #      / \                    / \
            #  ---*---o---      or    ---o---*---
            #     0\   \                /   /0
            #       \   *2            2*   /
            #        \ /                \ /
            #         *1                 *3
            face = np.roll(face, -v_on_face[0])
            if vertices_distances[face[1]] < 0.:
                p2, p3 = vertices[face[[2, 3]]]
                iright = plane.get_edge_intersection(p2, p3)
                intersections_vertices.append(iright)
                boundary_edge = [face[0], index_new_vertices]
                clipped_crown_mesh_faces.append([face[0], face[1], face[2], index_new_vertices])
            else:
                p1, p2 = vertices[face[[1, 2]]]
                ileft = plane.get_edge_intersection(p1, p2)
                intersections_vertices.append(ileft)
                boundary_edge = [index_new_vertices, face[0]]
                clipped_crown_mesh_faces.append([index_new_vertices, face[2], face[3], face[0]])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 1

        elif face_type == '013':
            # -----*-----
            #     / \
            #    /   \
            #   *     *
            #    \   /
            #     \ /
            #      *
            boundary_edge = None
            clipped_crown_mesh_faces.append(list(face))
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '210' or face_type == '310':
            #   *-------*               *
            #    \ 210 /               / \ 310
            #     \   /               *   *
            #      \ /                 \ /
            #   ----*----           ----*----
            boundary_edge = None

        elif face_type == '111':
            #        *2              *1
            #       /|               |\
            #      / |               | \
            #  ---*--o---    or   ---o--*---
            #     0\ |               | /0
            #       \|               |/
            #        *1              *2
            face = np.roll(face, -v_on_face[0])
            p1, p2 = vertices[face[[1, 2]]]
            if vertices_distances[face[1]] < 0.:
                iright = plane.get_edge_intersection(p1, p2)
                intersections_vertices.append(iright)
                boundary_edge = [face[0], index_new_vertices]
                clipped_crown_mesh_faces.append([face[0], face[1], index_new_vertices, face[0]])
            else:
                ileft = plane.get_edge_intersection(p1, p2)
                intersections_vertices.append(ileft)
                boundary_edge = [index_new_vertices, face[0]]
                clipped_crown_mesh_faces.append([index_new_vertices, face[2], face[0], index_new_vertices])
            clipped_crown_mesh_relative_faces_ids.append(face_id)
            index_new_vertices += 1

        elif face_type == '120':
            #         *O
            #        / \
            #       /   \
            #     1/     \2
            # ----*-------*----
            # face = np.roll(face, -v_above_face[0])
            # boundary_edge = [face[1], face[2]]
            # FIXME: quick fix here : robust ?
            boundary_edge = None

        elif face_type == '021':
            #  ----*-------*----
            #      2\     /1
            #        \   /
            #         \ /
            #          *0
            face = np.roll(face, -v_below_face[0])
            boundary_edge = [face[2], face[1]]
            face = list(face)
            face.append(face[0])
            clipped_crown_mesh_faces.append(face)
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '022':
            # ----*-----*----
            #    0|     |3
            #     |     |
            #    1*-----*2
            if v_on_face[1] == v_on_face[0] + 1:
                face = np.roll(face, -v_on_face[1])
            boundary_edge = [face[0], face[3]]
            clipped_crown_mesh_faces.append(list(face))
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '012':
            #   ------*------
            #        / \
            #       /   \
            #      /     \
            #     *-------*
            boundary_edge = None
            face = list(face)
            face.append(face[0])
            clipped_crown_mesh_faces.append(face)
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '220':
            #    0*-----*3
            #     |     |
            #    1|     |2
            # ----*-----*----

            # if v_above_face[1] == v_above_face[0] + 1:
            #     face = np.roll(face, -v_above_face[1])
            # boundary_edge = [face[1], face[2]]
            # FIXME: quick fix here : robust ?
            boundary_edge = None

        elif face_type == '121':
            #       *0
            #      / \
            #     /   \
            # ---*-----*---
            #    1\   /3
            #      \ /
            #       *2
            face = np.roll(face, -v_above_face[0])
            boundary_edge = [face[1], face[3]]
            clipped_crown_mesh_faces.append([face[1], face[2], face[3], face[1]])
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '300' or face_type == '400':
            #       *               *-----*
            #      / \              |     |
            #     /300\      or     | 400 |
            #    *-----*            *-----*
            # ____________       ______________
            boundary_edge = None

        elif face_type == '003':
            #  -----------
            #       *
            #      / \
            #     /   \
            #    *-----*
            boundary_edge = None
            face = list(face)
            face.append(face[0])
            clipped_crown_mesh_faces.append(face)
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '004':
            #  ---------------
            #      *-----*
            #      |     |
            #      |     |
            #      *-----*
            boundary_edge = None
            clipped_crown_mesh_faces.append(list(face))
            clipped_crown_mesh_relative_faces_ids.append(face_id)

        elif face_type == '030' or face_type == '040':
            # Face is totally on the plane --> rare case...
            boundary_edge = None

        else:
            raise Exception("Face %u clipping case %s not known." % (face_id, face_type))

        # Building boundary connectivity
        if boundary_edge is not None:
            direct_boundary_edges[boundary_edge[0]] = boundary_edge[1]
            inv_boundary_edges[boundary_edge[1]] = boundary_edge[0]

    if len(intersections_vertices) > 0:
        vertices = np.concatenate((vertices, intersections_vertices))

    clipped_crown_mesh = Mesh(vertices, clipped_crown_mesh_faces)
    clipped_crown_mesh._clipping_data = {
        'faces_ids': crown_mesh._clipping_data['faces_ids'][clipped_crown_mesh_relative_faces_ids]
    }

    return clipped_crown_mesh



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import capytaine as cpt


def print_matrix(x):
    for i in range(2):
        print(" %+19.12E %+19.12E %+19.12E %+19.12E" %
              (np.real(x[i, 0]), np.imag(x[i, 0]),
               np.real(x[i, 1]), np.imag(x[i, 1])))
    return


if __name__ == '__main__':

    print(' -- Run libdelhommeau/examples/minimal/minimal_example.py')

    vert = np.array([0.0, 0.0, -1.0,
                     1.0, 0.0, -1.0,
                     1.0, 1.0, -1.0,
                     0.0, 1.0, -1.0,
                     1.0, 0.0, -1.0,
                     2.0, 0.0, -1.0,
                     2.0, 1.0, -1.0,
                     1.0, 1.0, -1.0]).reshape((8, 3))

    faces = np.arange(0, 8).reshape(2, 4)
    mesh = cpt.Mesh(vert, faces)

    S, K = cpt.Delhommeau().evaluate(mesh, mesh, water_depth=np.inf, free_surface=np.inf)
    print(" Rankine part: S")
    print_matrix(S)
    print(" Rankine part: K")
    print_matrix(K)

    print(" k=1.0, h=inf: S")
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, water_depth=np.inf, wavenumber=1.0)
    print_matrix(S)
    print(" k=1.0, h=inf: K")
    print_matrix(K)

    S, K = cpt.Delhommeau().evaluate(mesh, mesh, water_depth=np.inf, wavenumber=2.0)
    print(" k=2.0, h=inf: S")
    print_matrix(S)
    print(" k=2.0, h=inf: K")
    print_matrix(K)

    print(" k=1.0, h=2.0: S")
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, water_depth=2., wavenumber=1.0)
    print_matrix(S)
    print(" k=1.0, h=2.0: K")
    print_matrix(K)

    S, K = cpt.Delhommeau().evaluate(mesh, mesh, water_depth=2., wavenumber=2.0)
    print(" k=2.0, h=2.0: S")
    print_matrix(S)
    print(" k=2.0, h=2.0: K")
    print_matrix(K)



#!/usr/bin/env python

import numpy as np
import capytaine as cpt
import capytaine.io.xarray

cpt.set_logging('INFO')

# Profile of the axisymmetric body
def shape(z):
    return 0.1*(-(z+1)**2 + 16)

# Generate the mesh and display it with VTK.
buoy = cpt.FloatingBody(
    mesh=cpt.AxialSymmetricMesh.from_profile(shape, z_range=np.linspace(-5, 0, 30), nphi=40)
)
buoy.add_translation_dof(name="Heave")
buoy.show()

# Set up problems
omega_range = np.linspace(0.1, 5.0, 60)
problems = [cpt.RadiationProblem(body=buoy, radiating_dof='Heave', omega=omega)
            for omega in omega_range]

# Solve the problems using the axial symmetry
solver = cpt.BEMSolver(engine=cpt.HierarchicalToeplitzMatrixEngine())
results = [solver.solve(pb) for pb in sorted(problems)]
dataset = capytaine.io.xarray.assemble_dataset(results)

# Plot results
import matplotlib.pyplot as plt
plt.figure()
plt.plot(
    omega_range,
    dataset['added_mass'].sel(radiating_dof='Heave',
                              influenced_dof='Heave'),
    label="Added mass",
)
plt.plot(
    omega_range,
    dataset['radiation_damping'].sel(radiating_dof='Heave',
                                     influenced_dof='Heave'),
    label="Radiation damping",
)
plt.xlabel('omega')
plt.grid()
plt.legend()
plt.show()



"""This module implements a class to describe a low-rank matrix as the tensor product of two smaller matrices.
In particular, an implementation of the Adaptive Cross Approximation is used to build such a matrix.

It takes inspiration from the following works:

* `openHmx module from Gypsilab by Matthieu Aussal (GPL licensed) <https://github.com/matthieuaussal/gypsilab>`_
* `HierarchicalMatrices by Markus Neumann (GPL licensed) <https://github.com/maekke97/HierarchicalMatrices>`_
"""
# Copyright (C) 2017-2021 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np

LOG = logging.getLogger(__name__)


class NoConvergenceOfACA(Exception):
    pass


class LowRankMatrix:
    """Matrix defined as the tensor product of two small matrices.

    Parameters
    ----------
    left_matrix: numpy.array
        Matrix of shape (nb_cols, rank).
    right_matrix: numpy.array
        Matrix of shape (rank, nb_rows).

    Attributes
    ----------
    shape: Tuple[int, int]
        The shape of the full matrix.
    rank: int
        The rank of the full matrix.
    dtype: numpy.dtype
        Type of data in the matrix.
    """

    ndim = 2

    ##############
    #  Creation  #
    ##############

    def __init__(self, left_matrix, right_matrix):
        self.left_matrix = left_matrix
        self.right_matrix = right_matrix
        self.shape = left_matrix.shape[0], right_matrix.shape[1]
        assert left_matrix.shape[1] == right_matrix.shape[0], "Sizes of the left and right matrices do not match."
        self.rank = left_matrix.shape[1]  # == right_matrix.shape[0]
        assert left_matrix.dtype == right_matrix.dtype, "Left and right matrices should have the same type of data."
        self.dtype = left_matrix.dtype

    @classmethod
    def from_full_matrix_with_SVD(cls, full_matrix, max_rank):
        """Create a low rank matrix from a full matrix using Singular Value Decomposition.

        Parameters
        ----------
        full_matrix: numpy array
            The matrix that will be approximated.
        max_rank: int
            Rank of the low-rank approximation.

        Returns
        -------
        LowRankMatrix
        """
        u, s, v = np.linalg.svd(full_matrix)
        left_matrix = u[:, 0:max_rank] @ np.diag(s[0:max_rank])
        right_matrix = v[0:max_rank, :]
        return cls(left_matrix, right_matrix)

    @classmethod
    def from_full_matrix_with_ACA(cls, full_matrix, max_rank=None, tol=0.0):
        """Create a low rank matrix from a full matrix using Adaptive Cross Approximation.
        The user should provide either the `max_rank` optional argument or the `tol` optional argument.

        Parameters
        ----------
        full_matrix: numpy.array
            The matrix that will be approximated.
        max_rank: int, optional
            The maximum rank allowed for the output low rank matrix.
            The default value is half the size of the full matrix, that is no gain in storage space.
        tol: float, optional
            The tolerance on the relative error (default: 0).
            If the Frobenius norm of the increment is lower than the tolerance, the iteration stops.
            If the tolerance is set to 0, the resulting matrix will have the maximum rank defined by `max_rank`.

        Returns
        -------
        LowRankMatrix
        """
        def get_row(i):
            return full_matrix[i, :]

        def get_col(j):
            return full_matrix[:, j]

        return cls.from_rows_and_cols_functions_with_ACA(
            get_row, get_col, full_matrix.shape[0], full_matrix.shape[1],
            max_rank=max_rank, tol=tol, dtype=full_matrix.dtype
        )

    @classmethod
    def from_function_with_ACA(cls, func, nb_rows, nb_cols, max_rank=None, tol=0.0, dtype=np.float64):
        """Create a low rank matrix from a function using Adaptive Cross Approximation.
        The user should provide either the `max_rank` optional argument or the `tol` optional argument.

        Parameters
        ----------
        func: Function
            Function such that `func(i, j)` returns the value of the (i, j) entry of the full matrix.
        nb_rows: int
            Number of rows in the full matrix.
        nb_cols: int
            Number of cols in the full matrix.
        max_rank: int, optional
            The maximum rank allowed for the output low rank matrix.
            The default value is half the size of the full matrix, that is no gain in storage space.
        tol: float, optional
            The tolerance on the relative error (default: 0).
            If the Frobenius norm of the increment is lower than the tolerance, the iteration stops.
            If the tolerance is set to 0, the resulting matrix will have the maximum rank defined by `max_rank`.
        dtype: numpy.dtype, optional
            Type of the data returned by the function.

        Returns
        -------
        LowRankMatrix
        """
        def get_row(i):
            return np.asarray([func(i, j) for j in range(nb_cols)])

        def get_col(j):
            return np.asarray([func(i, j) for i in range(nb_rows)])

        return cls.from_rows_and_cols_functions_with_ACA(
            get_row, get_col, nb_rows, nb_cols,
            max_rank=max_rank, tol=tol, dtype=dtype
        )

    @classmethod
    def from_rows_and_cols_functions_with_ACA(cls, get_row_func, get_col_func, nb_rows, nb_cols, max_rank=None, tol=0.0, dtype=np.float64):
        """Create a low rank matrix from functions using Adaptive Cross Approximation.
        The user should provide either the `max_rank` optional argument or the `tol` optional argument.

        Parameters
        ----------
        get_row_func: Function
            Function such that `get_row_func(i)` returns the `i`-th row of the full matrix.
        get_col_func: Function
            Function such that `get_col_func(j)` returns the `j`-th column of the full matrix.
        nb_rows: int
            Number of rows in the full matrix.
        nb_cols: int
            Number of columns in the full matrix.
        max_rank: int, optional
            The maximum rank allowed for the output low rank matrix.
            The default value is half the size of the full matrix, that is no gain in storage space.
        tol: float, optional
            The tolerance on the relative error (default: 0).
            If the Frobenius norm of the increment is lower than the tolerance, the iteration stops.
            If the tolerance is set to 0, the resulting matrix will have the maximum rank defined by `max_rank`.
        dtype: numpy.dtype, optional
            The type of data in the low rank matrix (default: float64).

        Returns
        -------
        LowRankMatrix
        """
        # Just some wrapping and unwrapping to use the multi-ACA below.
        def get_row(i):
            return [get_row_func(i)]

        def get_col(j):
            return [get_col_func(j)]

        return cls.from_rows_and_cols_functions_with_multi_ACA(
            get_row, get_col, nb_rows, nb_cols,
            nb_matrices=1, id_main=0,
            max_rank=max_rank, tol=tol, dtype=dtype
        )[0]

    @classmethod
    def from_rows_and_cols_functions_with_multi_ACA(cls, get_row, get_col, nb_rows, nb_cols,
                                                    nb_matrices=1, id_main=0,
                                                    max_rank=None, tol=0.0, dtype=np.float64):
        """Create several low rank matrices while running an Adaptive Cross Approximation.
        The user should provide either the `max_rank` optional argument or the `tol` optional argument.

        In Capytaine, the routines evaluating the influence matrices return the values of two matrices
        (S and V) at once, because there is a lot of common computations in their evaluation.
        The present function can be used to build the ACA of one of them while getting at the same time
        an approximation of the other for free.

        Freely adapted from the routine hmxACA.m from Gypsilab.

        Parameters
        ----------
        get_row: Function
            Function such that `get_row(i)` returns the `i`-th row of all the full matrices.
        get_col: Function
            Function such that `get_col(j)` returns the `j`-th column of all the full matrices.
        nb_rows: int
            Number of rows in all full matrices.
        nb_cols: int
            Number of columns in all full matrices.
        nb_matrices: int, optional
            The number of matrices approximated at the same time.
        id_main: int, optional
            The matrix used primarily in the ACA.
        max_rank: int, optional
            The maximum rank allowed for both output low rank matrices.
            The default value is half the size of the full matrix, that is no gain in storage space.
        tol: float, optional
            The tolerance on the relative error (default: 0).
            If the Frobenius norm of the increment is lower than the tolerance, the iteration stops.
            If the tolerance is set to 0, the resulting matrix will have the maximum rank defined by `max_rank`.
        dtype: numpy.dtype, optional
            The type of data in both low rank matrices (default: float64).

        Returns
        -------
        List[LowRankMatrix]
        """
        if max_rank is None and tol <= 0.0:
            LOG.warning("No stopping criterion for the Adaptive Cross Approximation."
                        "Please provide either max_rank or tol.")

        if max_rank is None:
            max_rank = min(nb_rows, nb_cols)//2

        # Initialize work matrices
        left = np.zeros((nb_matrices, nb_rows, max_rank), dtype=dtype)
        right = np.zeros((nb_matrices, max_rank, nb_cols), dtype=dtype)

        squared_norm_of_low_rank_approximation = 0.0

        # List of indices of unused entries in the full matrix
        available_rows = list(range(nb_rows))
        available_cols = list(range(nb_cols))

        for l in range(max_rank):
            # Pick a row
            if l == 0:
                relative_i = 0
                # Could also have been chosen at random.
            else:
                relative_i = int(np.argmax(np.abs(left[id_main, available_rows, l-1])))
                # The "int" is useless except for my type checker...

            i = available_rows.pop(relative_i)
            # relative_i is the index of the row in the list of remaining rows,
            # e.g. if available_rows = [2, 7, 8] and relative_i = 2, the chosen
            # row has index i = 8 in the full matrix.

            # Add the chosen row to the approximation of all the matrices
            one_row = get_row(i)
            for id_mat in range(nb_matrices):
                right[id_mat, l, :] = one_row[id_mat] - left[id_mat, i, :l] @ right[id_mat, :l, :]

            # Pick a column
            relative_j = int(np.argmax(np.abs(right[id_main, l, available_cols])))
            j = available_cols.pop(relative_j)
            # Similar to i above.

            one_col = get_col(j)

            # Add the column to the approximations of all matrices.
            for id_mat in range(nb_matrices):
                new_col = one_col[id_mat] - left[id_mat, :, :l] @ right[id_mat, :l, j]
                pivot = new_col[i]
                if abs(pivot) < 1e-12:
                    pivot = 1e-12
                left[id_mat, :, l] = new_col/pivot

            # Update norm of the full matrix
            squared_norm_of_increment = np.real(
                    (np.conj(left[id_main, :, l]) @ left[id_main, :, l]) *
                    (np.conj(right[id_main, l, :]) @ right[id_main, l, :])
            )

            crossed_terms = (
                    (np.conj(left[id_main, :, l].T) @ left[id_main, :, :l]) @
                    (np.conj(right[id_main, l, :]) @ right[id_main, :l, :].T)
            )
            squared_norm_of_low_rank_approximation += squared_norm_of_increment + 2*np.real(crossed_terms)

            if squared_norm_of_increment <= tol**2*squared_norm_of_low_rank_approximation:
                LOG.debug(f"The ACA has found an approximation of rank {l}.")

                if l == 0:  # Edge case of the zero matrix, ...
                    l = 1  # ... we actually return a "rank 1" LowRankMatrix with coefficients equal to zero.

                return [LowRankMatrix(left[id_mat, :, :l], right[id_mat, :l, :]) for id_mat in range(nb_matrices)]

        if tol > 0:
            LOG.warning(f"The ACA was unable to find a low rank approximation "
                        f"of rank lower or equal to {max_rank} with tolerance {tol:.2e} (latest iteration: "
                        f"{np.sqrt(squared_norm_of_increment):.2e}/{np.sqrt(squared_norm_of_low_rank_approximation):.2e}).")
            raise NoConvergenceOfACA()

        return [LowRankMatrix(left[id_mat, :, :], right[id_mat, :, :]) for id_mat in range(nb_matrices)]

    ####################
    #  Representation  #
    ####################

    def full_matrix(self, dtype=None):
        if dtype is not None:
            return (self.left_matrix @ self.right_matrix).astype(dtype)
        else:
            return self.left_matrix @ self.right_matrix

    def __array__(self, dtype=None):
        return self.full_matrix(dtype=dtype)

    @property
    def stored_data_size(self):
        return np.prod(self.left_matrix.shape) + np.prod(self.right_matrix.shape)

    @property
    def density(self):
        return self.stored_data_size/np.prod(self.shape)

    @property
    def sparcity(self):
        return 1 - self.density

    ####################
    #  Transformation  #
    ####################

    def recompress(self, tol=None, new_rank=None):
        """Recompress the matrix to a lower rank. Based on the routine hmxQRSVD.m from Gipsylab."""
        if new_rank is None:
            new_rank = self.rank
        QA, RA = np.linalg.qr(self.left_matrix)
        QB, RB = np.linalg.qr(self.right_matrix.T)
        U, S, V = np.linalg.svd(RA @ RB.T)
        if tol is not None:
            new_rank = np.count_nonzero(S/S[0] >= tol)
        A = QA @ (U[:, :new_rank] @ np.diag(S[:new_rank]))
        B = QB @ V[:, :new_rank]
        return LowRankMatrix(A, B.T)

    def __add__(self, other):
        if isinstance(other, LowRankMatrix):
            new_left = np.concatenate([self.left_matrix, other.left_matrix], axis=1)
            new_right = np.concatenate([self.right_matrix, other.right_matrix], axis=0)
            return LowRankMatrix(new_left, new_right).recompress(new_rank=min(self.rank, other.rank))
        else:
            return NotImplemented

    def __neg__(self):
        return LowRankMatrix(-self.left_matrix, self.right_matrix)

    def __sub__(self, other):
        return self + (-other)

    def __truediv__(self, other):
        from numbers import Number
        if isinstance(other, Number):
            return LowRankMatrix(self.left_matrix, self.right_matrix/other)
        else:
            return NotImplemented

    def __matmul__(self, other):
        if isinstance(other, np.ndarray) and len(other.shape) == 1:
            return self._mul_with_vector(other)
        else:
            return NotImplemented

    def _mul_with_vector(self, other):
        return self.left_matrix @ (self.right_matrix @ other)

    def __rmatmul__(self, other):
        if isinstance(other, np.ndarray) and len(other.shape) == 1:
            return self._rmul_with_vector(other)
        else:
            return NotImplemented

    def _rmul_with_vector(self, other):
        return (other @ self.left_matrix) @ self.right_matrix

    def astype(self, dtype):
        return LowRankMatrix(self.left_matrix.astype(dtype), self.right_matrix.astype(dtype))
